import os

from cloudmesh.common.util import readfile


def run(command):
    """
    run the command, redirect the outout to a file and display the content once
    it is completed.

    :param command:
    :return:
    """
    os.system(f"{command} &> ./cmd-output")
    content = readfile("./cmd-output")
    return content



import subprocess


def run(command, shell=True):
    """
    runs the command and returns the output in utf-8 format

    :param command:
    :param shell:
    :return:
    """
    result = subprocess.check_output(command, shell=shell)
    return result.decode("utf-8")



import json


class TableParser(object):
    @classmethod
    def convert(cls, table=None,
                output='dict',
                header=True,
                index=None,
                change=None,
                strip=True,
                lower=True,
                strip_seperator=True,
                seperator="|",
                comment_chars="+#"):
        if change is None:
            change = [(":", "_"), ("(", "_"), (")", ""), ("/", "_")]
        parser = TableParser(
            output=output,
            header=header,
            index=index,
            change=change,
            strip=strip,
            lower=lower,
            strip_seperator=strip_seperator,
            seperator=seperator,
            comment_chars=comment_chars)
        if table is not None:
            if 'dict' in output:
                return parser.to_dict(table)
            elif 'list' in output:
                return parser.to_dict(table)
            else:
                raise ValueError("output type not supported")

    def __init__(self,
                 output='dict',
                 header=True,
                 index=None,
                 change=None,
                 strip=True,
                 lower=True,
                 strip_seperator=True,
                 seperator="|",
                 comment_chars="+#"):
        """

        :param header: if true the first line is a header. Not implemented
        :param index: if true, identifies one of the header column as index
                      for dict key naming
        :param change: an array of tuples of characters that need to be
                       changed to allow key creation in the dict
        :param strip: if true the the lines start and end with the separator
        :param lower: converts headers to lower case
        :param strip_seperator: removes the spaces before and after the
                                separator
        :param seperator: the separator character, default is |
        :param comment_chars: lines starting with this chars will be ignored
        :return:
        """
        if change is None:
            change = [(":", "_"), ("(", "_"), (")", ""), ("/", "_")]
        self.data = None
        self.output = None
        self.with_header = header
        self.change = change
        self.is_lower = lower
        self.is_strip = strip
        self.index = index
        self.strip_seperator = strip_seperator
        self.seperator = seperator
        self.comment_chars = comment_chars
        self.lines = None

    def clean(self, line):
        """
        :param line: cleans the string
        :return:
        """
        # print ("-" + line + "-")
        if line == '':
            line = 'None'
        if self.is_lower:
            line = line.lower()
        if line == "user ":  # for slurm which has "user" and "user "
            line = "userid"
        for convert in self.change:
            line = line.replace(convert[0], convert[1])
        if self.is_strip:
            line = line.strip()
        return line.strip(' ')

    def extract_lines(self, table):
        lines = table.splitlines()
        self.lines = []
        for line in lines:
            if self.is_strip:
                line = line.strip()
            if line[0] not in self.comment_chars:
                self.lines.append(line)
        return self.lines

    def _get_headers(self):
        """
        assumes comment have been stripped with extract
        :return:
        """

        header = self.lines[0]
        self.lines = self.lines[1:]

        self.headers = \
            [self.clean(h) for h in header.split(self.seperator)]
        if self.is_strip:
            self.headers = self.headers[1:-1]
        return self.headers

    def to_dict(self, table):
        """

        :param table: converts a stream called line to a dict
        :type table: string
        :return: the dict
        """

        self.extract_lines(table)

        self._get_headers()

        i = 0
        self.data = {}
        for line in self.lines:
            element = [h.strip() for h in line.split(self.seperator)]
            if self.is_strip:
                element = element[1:-1]
            entry = {}
            for column in range(0, len(self.headers)):
                entry[self.headers[column]] = element[column]
            if self.index is None:
                self.data[str(i)] = dict(entry)
            else:
                self.data[entry[self.index]] = dict(entry)
            i += 1
        return self.data

    def to_list(self, table):
        """
        :param table: converts a stream called line to a list
        :type table: string
        :return: the list
        """
        self.extract_lines(table)

        self._get_headers()

        i = 0
        self.data = []
        for line in self.lines:
            element = [h.strip() for h in line.split(self.seperator)]
            if self.is_strip:
                element = element[1:-1]

            entry = {}
            for column in range(0, len(self.headers)):
                entry[self.headers[column]] = element[column]
            if self.index is None:
                entry["_id"] = str(i)
            else:
                entry["_id"] = self.index
            self.data.append(entry)
            i += 1
        return self.data

    def json(self):
        return self.data

    def __str__(self):
        return json.dumps(self.data, indent=4, separators=(',', ': '))


if __name__ == "__main__":
    parser = TableParser()
    d = parser.to_list("|a|b|c|\n|1|2|3|\n+|4|5|6|\n|7|8|9|")
    print(d)
    print(parser.json())
    print(parser.headers)
    print(parser)

    parser.to_dict("|a|b|c|\n|1|2|3|\n|4|5|6|")
    print(parser.json())

    print(parser.headers)
    print(parser)



import textwrap

from cloudmesh.common.JobSet import JobSet
from cloudmesh.common.Shell import Shell
from cloudmesh.common.console import Console
from cloudmesh.common.parameter import Parameter


class JobMultiHostScript:
    '''
    The JobMultiHostScript is a simple mechanism to run a number of commands formylated
    in a script in parallel over the hosts provided. The script is interpreted line
    by line and does not support multi line commands at this time. (Not difficult
    to implement when looking at \\ at the end of a line.)

    Static method::
        script = """
          # This is a comment
          pwd     # tag: pwd
          uname -a
        """

        hosts = Parameter.expand("purple[01-02]")
        result = JobMultiHostScript.execute(script, "script_name", hosts)
    '''

    def __init__(self, name, script, hosts, executor):
        self.name = name
        self.script = script
        self.hosts = hosts
        self.executor = executor

    def run(self, beginLine=None, endLine=None):
        # Prepare script
        self.script = textwrap.dedent(str(self.script))
        lines = self.script.splitlines()

        # Truncate lines if begin and/or line parameters are set
        if beginLine is not None and endLine is not None:
            lines = Shell.find_lines_between(lines, beginLine, endLine)
        elif beginLine is not None and endLine is None:
            lines = Shell.find_lines_from(lines, beginLine)
        elif beginLine is None and endLine is not None:
            lines = Shell.find_lines_to(lines, endLine)

        # Loop over each line
        for line in lines:
            stripped = line.strip()
            # Check if line is a comment
            if stripped.startswith("#") or stripped == "":
                pass
            else:
                # Create jobSet per line
                job = JobSet(self.name, executor=self.executor)
                if "# tag:" in line:
                    line, tag = line.split("# tag:", 1)
                    tag = tag.strip()
                    line = line.strip()

                # Execute jobSet for each host in parallel
                for host in self.hosts:
                    job.add_directory({"name": host, "host": host, "command": line})
                job.run(parallel=len(self.hosts))
                job.Print()

    @staticmethod
    def execute(script, name="script", hosts=None, executor=JobSet.ssh,
                beginLine=None, endLine=None):
        job = JobMultiHostScript(name, script, hosts, executor)
        job.run(beginLine, endLine)

    # CMS Function
    def cms(self, arguments):
        # Prepare parameters
        script = None
        if arguments.script:
            script = arguments.script

        hosts = None
        if arguments.hosts:
            hosts = arguments.hosts

        name = "script"
        if arguments.name:
            name = name

        executor = JobSet.ssh
        if arguments.executor:
            executor = arguments.executor

        beginLine = None
        if arguments.begin:
            beginLine = arguments.begin

        endLine = None
        if arguments.end:
            endLine = arguments.end

        # Execute
        if script is not None and hosts is not None:
            job = JobMultiHostScript(name, script, hosts, executor)
            job.run(beginLine, endLine)
        else:
            if hosts is None:
                Console.error("The hosts are not defined, found None as content")
            if script is None:
                Console.error("The script is not defined, found None as content")
            return


if __name__ == '__main__':
    script = """
    # This is a comment

    # Task: pwd
    pwd     # tag: pwd

    # Task: uname
    uname -a
    """

    hosts = Parameter.expand("purple[01-02]")
    result = JobMultiHostScript.execute(script, "script_name", hosts,
                                        beginLine="# Task: pwd", endLine="# Task: uname")



import datetime as TIME
import humanize as HUMANIZE
from dateutil import parser
import calendar


class DateTime(object):
    """
    This class provides some simple date time functions so we can use all the
    same format. Here is a simple example

        start = DateTime.now()
        stop = DateTime.now() + DateTime.delta(1)

        print ("START", start)
        print ("STOP", stop)
        print("HUMANIZE STOP", DateTime.humanize(stop - start))
        print ("LOCAL", DateTime.local(start))
        print("UTC", DateTime.utc(start))
        print("NATURAL", DateTime.natural(start))
        print("WORDS", DateTime.words(start))
        print("TIMEZONE", DateTime.timezone)

    This will result in

        START 2019-08-03 21:34:14.019147
        STOP 2019-08-03 21:34:15.019150
        HUMANIZE STOP a second ago
        LOCAL 2019-08-03 17:34:14 EST
        UTC 2019-08-03 21:34:14.019147 UTC
        NATURAL 2019-08-03 21:34:14.019147 UTC
        WORDS Sat 6 Aug 2019, 21:34:14 UTC
        TIMEZONE EST

    """

    # timezone = time.tzname[0]

    @staticmethod
    def now():
        return TIME.datetime.utcnow()

    @staticmethod
    def natural(time):
        if type(time) == TIME.datetime:
            time = str(time)
        return str(parser.parse(time)) + " UTC"

    @staticmethod
    def words(time):
        return TIME.datetime.strftime(time, "%a %w %b %Y, %H:%M:%S UTC")

    @staticmethod
    def datetime(time):
        if type(time) == TIME:
            return time
        else:
            return DateTime.humanize(time)
        pass

    @staticmethod
    def humanize(time):
        return HUMANIZE.naturaltime(time)

    @staticmethod
    def string(time):
        if type(time) == str:
            d = parser.parse(time)
        else:
            d = time
        return d

    @staticmethod
    def delta(n):
        return TIME.timedelta(seconds=n)

    @staticmethod
    def utc(time):
        return str(time) + " UTC"

    @staticmethod
    def local(time):
        return DateTime.utc_to_local(time)

    def utc_to_local(time):
        TIME_FORMAT = '%Y-%m-%d %H:%M:%S'
        utc = TIME.datetime.utcnow().strftime(TIME_FORMAT)
        timestamp = calendar.timegm(
            (TIME.datetime.strptime(utc, TIME_FORMAT)).timetuple())
        local = TIME.datetime.fromtimestamp(timestamp).strftime(TIME_FORMAT)
        return local + " " + DateTime.timezone


if __name__ == "__main__":
    start = DateTime.now()
    stop = DateTime.now() + DateTime.delta(1)

    print("START", start)
    print("STOP", stop)
    print("HUMANIZE STOP", DateTime.humanize(stop - start))
    print("LOCAL", DateTime.local(start))
    print("UTC", DateTime.utc(start))
    print("NATURAL", DateTime.natural(start))
    print("WORDS", DateTime.words(start))
    print("TIMEZONE", DateTime.timezone)

    # print("CONVERT", DateTime.local("2019-08-03 20:48:27.205442 UTC"))
    """
    START 2019-08-03 21:34:14.019147
    STOP 2019-08-03 21:34:15.019150
    HUMANIZE STOP a second ago
    LOCAL 2019-08-03 17:34:14 EST
    UTC 2019-08-03 21:34:14.019147 UTC
    NATURAL 2019-08-03 21:34:14.019147 UTC
    WORDS Sat 6 Aug 2019, 21:34:14 UTC
    TIMEZONE EST
    """



"""
A convenient method to execute shell commands and return their output. Note:
that this method requires that the command be completely execute before the
output is returned. FOr many activities in cloudmesh this is sufficient.

"""
import errno
import glob
import os
import platform as os_platform
import requests
import shlex
import shutil
import subprocess
import sys
import textwrap
import webbrowser
import zipfile
from pathlib import Path
from pipes import quote
from sys import platform
from tqdm import tqdm

import psutil
from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.console import Console
from cloudmesh.common.dotdict import dotdict
from cloudmesh.common.systeminfo import get_platform
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from cloudmesh.common.util import writefile
from cloudmesh.common.systeminfo import os_is_windows
from cloudmesh.common.systeminfo import os_is_mac
from cloudmesh.common.systeminfo import os_is_linux
from cloudmesh.common.util import is_gitbash


# from functools import wraps
# def timer(func):
#    @wraps(func)
#    def wrapper(*args,**kwargs):
#        print(f"{func.__name__!r} begins")
#        start_time = time.time()
#        result = func(*args,**kwargs)
#        print(f"{func.__name__!r} ends in {time.time()-start_time}  secs")
#        return result
#    return wrapper

# def NotImplementedInWindows(f):
#    def new_f(*args):
#        if sys.platform == "win32":
#            Console.error("The method {f.__name__} is not implemented in Windows,"
#                        " please implement, and/or submit an issue.")
#            sys.exit()
#        f(args)
#    return new_f

def windows_not_supported(f):
    def wrapper(*args, **kwargs):
        host = get_platform()
        if host == "windows":
            Console.error("Not supported on windows")
            return ""
        else:
            return f(*args, **kwargs)

    return wrapper


def NotImplementedInWindows():
    if sys.platform == "win32":
        Console.error(f"The method {__name__} is not implemented in Windows.")
        sys.exit()


class Brew(object):

    @classmethod
    def install(cls, name):

        r = Shell.brew("install", name)
        print(r)

        if "already satisfied: " + name in r:
            print(name, "... already installed")
            Console.ok(r)
        elif "Successfully installed esptool":
            print(name, "... installed")
            Console.ok(r)
        else:
            print(name, "... error")
            Console.error(r)

    @classmethod
    def version(cls, name):
        r = Shell.brew("ls", "--version", "name")
        print(r)


class Pip(object):
    @classmethod
    def install(cls, name):
        r = Shell.pip("install", name)
        if f"already satisfied: {name}" in r:
            print(name, "... already installed")
            Console.ok(r)
        elif f"Successfully installed {name}":
            print(name, "... installed")
            Console.ok(r)
        else:
            print(name, "... error")
            Console.error(r)


class SubprocessError(Exception):
    """
    Manages the formatting of the error and stdout.
    This command should not be directly called. Instead use Shell
    """

    def __init__(self, cmd, returncode, stderr, stdout):
        """
        sets the error
        :param cmd: the command executed
        :param returncode: the return code
        :param stderr: the stderr
        :param stdout: the stdout
        """
        self.cmd = cmd
        self.returncode = returncode
        self.stderr = stderr
        self.stdout = stdout

    def __str__(self):

        def indent(lines, amount, ch=' '):
            """
            indent the lines by multiples of ch
            :param lines:
            :param amount:
            :param ch:
            :return:
            """
            padding = amount * ch
            return padding + ('\n' + padding).join(lines.split('\n'))

        cmd = ' '.join(map(quote, self.cmd))
        s = ''
        s += 'Command: %s\n' % cmd
        s += 'Exit code: %s\n' % self.returncode

        if self.stderr:
            s += 'Stderr:\n' + indent(self.stderr, 4)
        if self.stdout:
            s += 'Stdout:\n' + indent(self.stdout, 4)

        return s


class Subprocess(object):
    """
    Executes a command. This class should not be directly used, but
    instead you should use Shell.
    """

    def __init__(self, cmd, cwd=None, stderr=subprocess.PIPE,
                 stdout=subprocess.PIPE, env=None):
        """
        execute the given command
        :param cmd: the command
        :param cwd: the directory in which to execute the command
        :param stderr: the pipe for stderror
        :param stdout: the pipe for the stdoutput
        :param env:
        """
        Console.debug_msg('Running cmd: {}'.format(' '.join(map(quote, cmd))))

        proc = subprocess.Popen(cmd, stderr=stderr, stdout=stdout, cwd=cwd,
                                env=env)
        stdout, stderr = proc.communicate()

        self.returncode = proc.returncode
        self.stderr = stderr
        self.stdout = stdout

        if self.returncode != 0:
            raise SubprocessError(cmd, self.returncode, self.stderr,
                                  self.stdout)


class Shell(object):
    """
    The shell class allowing us to conveniently access many operating system commands.
    TODO: This works well on Linux and OSX, but has not been tested much on Windows
    """

    # TODO: we have not supported cygwin for a while
    # cygwin_path = 'bin'  # i copied fom C:\cygwin\bin

    command = {
        'windows': {},
        'linux': {},
        'darwin': {}
    }

    # TODO
    #
    # how do we now define dynamically functions based on a list that we want to support
    #
    # what we want is where args are multiple unlimited parameters to the function
    #
    # def f(args...):
    #     name = get the name from f
    #     a = list of args...
    #
    #    cls.execute(cmd, arguments=a, capture=True, verbose=False)
    #
    # commands = ['ps', 'ls', ..... ]
    # for c in commands:
    #    generate this command and add to this class dynamically
    #
    # or do something more simple
    #
    # ls = cls.execute('cmd', args...)
    @staticmethod
    def timezone(default="America/Indiana/Indianapolis"):
        # BUG we need to be able to pass the default from the cmdline
        host = get_platform()
        if host == "windows":
            return default
        else:
            # result = Shell.run("ls -l /etc/localtime").strip().split("/")
            try:
                result = Shell.run("ls -l /etc/localtime").strip().split("zoneinfo")[1][1:]
                return result
            except IndexError as e:
                return default

    @staticmethod
    @windows_not_supported
    def locale():
        try:
            result = Shell.run('locale').split('\n')[0].split('_')[1].split('.')[0].lower()
            return result
        except IndexError as e:
            Console.warning('Could not determine locale. Defaulting to "us"')
            return 'us'

    @staticmethod
    def ssh_enabled():
        if os_is_linux():
            r = Shell.run("service sshd status | fgrep running").strip()
            return len(r) > 0
        elif os_is_windows():
            r = Shell.run("ps | grep -F ssh")
            return "ssh" in r
        elif os_is_mac():
            r = Shell.run("ps -ef")
            if "sshd" in r:
                print('IT WORKS!!!!!')
            else:
                print('it doenst work :(')
            return "sshd" in r
        return False

    @staticmethod
    def run_timed(label, command, encoding=None, service=None):
        """
        runs the command and uses the StopWatch to time it
        :param label: name of the StopWatch
        :param command: the command to be executed
        :param encoding: the encoding
        :param service: a prefix to the stopwatch label
        :return:
        """
        _label = str(label)
        print(_label, command)
        StopWatch.start(f"{service} {_label}")
        result = Shell.run(command, encoding)
        StopWatch.stop(f"{service} {_label}")
        return str(result)

    @staticmethod
    def run(command, exit="; exit 0", encoding='utf-8', replace=True, timeout=None):
        """
        executes the command and returns the output as string
        :param command:
        :param encoding:
        :return:
        """

        if sys.platform == "win32":
            if replace:
                c = "&"
            else:
                c = ";"
            command = f"{command}".replace(";", c)
        else:
            command = f"{command} {exit}"

        if timeout is not None:
            r = subprocess.check_output(command,
                                        stderr=subprocess.STDOUT,
                                        shell=True,
                                        timeout=timeout)
        else:
            r = subprocess.check_output(command,
                                        stderr=subprocess.STDOUT,
                                        shell=True)
        if encoding is None or encoding == 'utf-8':
            return str(r, 'utf-8')
        else:
            return r

    @staticmethod
    def run2(command, encoding='utf-8'):
        """
        executes the command and returns the output as string. This command also
        allows execution of 32 bit commands.

        :param command: the program or command to be executed
        :param encoding: encoding of the output
        :return:
        """
        if platform.lower() == 'win32':
            import ctypes

            class disable_file_system_redirection:
                _disable = ctypes.windll.kernel32.Wow64DisableWow64FsRedirection
                _revert = ctypes.windll.kernel32.Wow64RevertWow64FsRedirection

                def __enter__(self):
                    self.old_value = ctypes.c_long()
                    self.success = self._disable(ctypes.byref(self.old_value))

                def __exit__(self, type, value, traceback):
                    if self.success:
                        self._revert(self.old_value)

            with disable_file_system_redirection():
                command = f"{command}"
                r = subprocess.check_output(command,
                                            stderr=subprocess.STDOUT,
                                            shell=True)
                if encoding is None or encoding == 'utf-8':
                    return str(r, 'utf-8')
                else:
                    return r
        elif platform.lower() == 'linux' or platform.lower() == 'darwin':
            command = f"{command}"
            r = subprocess.check_output(command,
                                        stderr=subprocess.STDOUT,
                                        shell=True)
            if encoding is None or encoding == 'utf-8':
                return str(r, 'utf-8')
            else:
                return r

    @classmethod
    def execute(cls,
                cmd,
                arguments="",
                shell=False,
                cwd=None,
                traceflag=True,
                witherror=True):
        """Run Shell command

        :param witherror: if set to False the error will not be printed
        :param traceflag: if set to true the trace is printed in case of an error
        :param cwd: the current working directory in which the command is
        supposed to be executed.
        :param shell: if set to true the subprocess is called as part of a shell
        :param cmd: command to run
        :param arguments: we do not know yet
        :return:
        """
        # print "--------------"
        result = None
        terminal = cls.terminal_type()
        # print cls.command
        os_command = [cmd]
        if terminal in ['linux', 'windows']:
            os_command = [cmd]
        elif 'cygwin' in terminal:
            if not cls.command_exists(cmd):
                print("ERROR: the command could not be found", cmd)
                return
            else:
                os_command = [cls.command[cls.operating_system()][cmd]]

        if isinstance(arguments, list):
            os_command = os_command + arguments
        elif isinstance(arguments, tuple):
            os_command = os_command + list(arguments)
        elif isinstance(arguments, str):
            os_command = os_command + arguments.split()
        else:
            print("ERROR: Wrong parameter type", type(arguments))

        if cwd is None:
            cwd = os.getcwd()
        # noinspection PyPep8
        try:
            if shell:
                if platform.lower() == 'win32':
                    import ctypes

                    class disable_file_system_redirection:
                        _disable = ctypes.windll.kernel32.Wow64DisableWow64FsRedirection
                        _revert = ctypes.windll.kernel32.Wow64RevertWow64FsRedirection

                        def __enter__(self):
                            self.old_value = ctypes.c_long()
                            self.success = self._disable(
                                ctypes.byref(self.old_value))

                        def __exit__(self, type, value, traceback):
                            if self.success:
                                self._revert(self.old_value)

                    if len(os_command) == 1:
                        os_command = os_command[0].split(' ')
                    with disable_file_system_redirection():
                        result = subprocess.check_output(os_command,
                                                         stderr=subprocess.STDOUT,
                                                         shell=True,
                                                         cwd=cwd)
                else:
                    result = subprocess.check_output(
                        os_command,
                        stderr=subprocess.STDOUT,
                        shell=True,
                        cwd=cwd)
            else:
                result = subprocess.check_output(
                    os_command,
                    # shell=True,
                    stderr=subprocess.STDOUT,
                    cwd=cwd)
        except:  # noqa: E722
            if witherror:
                Console.error("problem executing subprocess",
                              traceflag=traceflag)
        if result is not None:
            result = result.strip().decode()
        return result

    @staticmethod
    def oneline(script, seperator=" && "):
        """
        converts a script to one line command.
        THis is useful to run a single ssh command and pass a one line script.

        :param script:
        :return:
        """
        return seperator.join(textwrap.dedent(script).strip().splitlines())

    @staticmethod
    def is_root():
        """
        checks if the user is root

        :return: True if the user is root
        :rtype: boolean
        """
        username = subprocess.getoutput("whoami")
        return username == "root"

    @staticmethod
    def rmdir(top, verbose=False):
        """
        removes a directory

        :param top: removes the directory tree from the top
        :type top: str
        :param verbose: unused
        :type verbose: unused
        :return: void
        :rtype: void
        """
        p = Path(top)
        if not p.exists():
            return
        try:
            shutil.rmtree(path_expand(top))
        except Exception as e:
            print(e)
            pass
            # print(e.strerror)
            # assert False, f"Directory {top} could not be deleted"

    @staticmethod
    def dot2svg(filename, engine='dot'):
        data = {
            'engine': engine,
            'file': filename.replace(".dot", "")
        }
        command = "{engine} -Tsvg {file}.dot > {file}.svg".format(**data)
        print(command)
        os.system(command)

    @staticmethod
    def map_filename(name):
        pwd = os.getcwd()

        _name = str(name)
        result = dotdict()

        if os_is_windows():
            if "USERNAME" in os.environ:
                result.user = os.path.basename(os.environ["USERNAME"])
            else:
                Console.error("Could not detect the home directory")
        else:
            result.user = os.path.basename(os.environ["HOME"])
        result.host = "localhost"
        result.protocol = "localhost"

        if _name.startswith("http"):
            result.path = _name
            result.protocol = _name.split(':', 1)[0]
        elif _name.startswith("wsl:"):
            result.path = _name.replace("wsl:", "")
            # Abbreviations: replace ~ with home dir and ./ + / with pwd
            if result.path.startswith("~"):
                if os_is_linux():
                    result.path = result.path.replace("~", f"/mnt/c/home/{result.user}")
                else:
                    result.path = result.path.replace("~", f"/mnt/c/Users/{result.user}")
            elif not result.path.startswith("/"):
                if os_is_windows():
                    pwd = pwd.replace("C:", "/mnt/c").replace("\\", "/")
                result.path = pwd + "/" + result.path.replace("./", "")
            result.protocol = "cp"
            result.host = "wsl"
        elif _name.startswith("scp:"):
            # scp source destination
            try:
                result.scp, userhost, result.path = _name.split(":")
                result.user, result.host = userhost.split("@")
                result.protocol = "scp"
            except:  # noqa: E722
                Console.error("The format of the name is not supported: {name}")
        elif _name.startswith("rsync:"):
            try:
                result.scp, userhost, result.path = _name.split(":")
                result.user, result.host = userhost.split("@")
                result.protocol = "rsync"
            except:  # noqa: E722
                Console.error("The format of the name is not supported: {name}")
        elif _name.startswith(".") or _name.startswith("~"):
            result.path = path_expand(_name)
        elif _name[1] == ":":
            drive, path = _name.split(":", 1)
            if os_is_windows():
                result.path = path_expand(path)
            else:
                result.path = drive + ":" + path_expand(path)
            result.path = result.path.replace("/", "\\")
        else:
            result.path = path_expand(_name)

        return result

    @staticmethod
    def browser(filename=None):
        """
        :param filename:
        :return:
        """
        if not os.path.isabs(filename) and 'http' not in filename:
            filename = Shell.map_filename(filename).path
        webbrowser.open(filename, new=2)

    @staticmethod
    def fetch(filename=None, destination=None):
        _filename = Shell.map_filename(filename)

        content = None

        if _filename.path.startswith('http'):
            result = requests.get(_filename.path)
            content = result.text

        else:
            os.path.exists(_filename.path)
            content = readfile(_filename.path)

        if destination is not None:
            destination = Shell.map_filename(destination)
            writefile(destination, content)

        return content

    @staticmethod
    def terminal_title(name):
        """
        sets the title of the terminal

        :param name: the title
        :type name: str
        :return: void
        :rtype: void
        """
        return f'echo -n -e \"\033]0;{name}\007\"'

    @classmethod
    def terminal(cls, command='pwd', title=None, kind=None):
        """
        starts a terminal and executes the command in that terminal

        :param command: the command to be executed
        :type command: str
        :param title: the title
        :type title: str
        :param kind: for windows you can set "cmd", "powershell", or "gitbash"
        :type kind: str
        :return: void
        :rtype: void
        """
        # title nameing not implemented
        print(platform)
        if platform == 'darwin':
            label = Shell.terminal_title(title)

            os.system(
                f"osascript -e 'tell application \"Terminal\" to do script \"{command}\"'"
            )
        elif platform == "linux":  # for ubuntu running gnome
            dist = os_platform.linux_distribution()[0]
            linux_apps = {'ubuntu': 'gnome-terminal', 'debian': 'lxterminal'}
            os.system(f"{linux_apps[dist]} -e \"bash -c \'{command}; exec $SHELL\'\"")

        elif platform == "win32":
            if kind is None:
                if Path.is_dir(Path(r"C:\Program Files\Git")):
                    kind = "gitbash"
            kind = kind.lower()
            if kind == "gitbash":
                p = subprocess.Popen([r"C:\Program Files\Git\git-bash.exe",
                                      "-c",
                                      f"{command}"])
                return p.pid
            elif kind == "cmd":
                Console.error(f"Command not implemented for {kind}")
            elif kind == "powershell":
                Console.error(f"Command not implemented for {kind}")
            else:
                Console.error("Git bash is not found, please make sure it "
                              "is installed. Other terminals not supported.")
                raise NotImplementedError
        else:
            raise NotImplementedError

    @classmethod
    def live(cls, command, cwd=None):
        if cwd is None:
            cwd = os.getcwd()
        process = subprocess.Popen(shlex.split(command), cwd=cwd,
                                   stdout=subprocess.PIPE)
        result = b''
        while True:
            output = process.stdout.read(1)
            if output == b'' and process.poll() is not None:
                break
            if output:
                result = result + output
                sys.stdout.write(output.decode("utf-8"))
                sys.stdout.flush()
        rc = process.poll()
        data = dotdict({
            "status": rc,
            "text": output.decode("utf-8")
        })
        return data

    @classmethod
    def get_python(cls):
        """
        returns the python and pip version
        :return: python version, pip version
        """
        python_version = sys.version_info[:3]
        v_string = [str(i) for i in python_version]
        python_version_s = '.'.join(v_string)

        # pip_version = pip.__version__
        pip_version = Shell.pip("--version").split()[1]
        return python_version_s, pip_version

    @classmethod
    def check_output(cls, *args, **kwargs):
        """Thin wrapper around :func:`subprocess.check_output`
        """
        return subprocess.check_output(*args, **kwargs)

    @classmethod
    def ls(cls, directory=".", match=None):
        """
        executes ls with the given arguments
        :param args:
        :return: list
        """
        if match == None:
            files = os.listdir('.')
        else:
            files = [f for f in os.listdir('.') if re.match(match, f)]
        return files

    @classmethod
    # @NotImplementedInWindows
    def unix_ls(cls, *args):
        """
        executes ls with the given arguments
        :param args:
        :return:
        """
        return cls.execute('ls', args)

    @staticmethod
    def ps(short=False, attributes=None):
        """
        using psutil to return the process information pid, name and comdline,
        cmdline may be a list

        :return: a list of dicts of process information
        """
        found = []
        for proc in psutil.process_iter():
            try:
                if attributes is not None:
                    pinfo = proc.as_dict(attrs=attributes)
                elif short:
                    pinfo = proc.as_dict(attrs=['pid', 'name', 'cmdline', 'ppid', 'username',
                                                'status', 'create_time', 'terminal', 'cwd',
                                                'open_files'])
                else:
                    pinfo = proc.as_dict()
            except psutil.NoSuchProcess:
                pass
            else:
                found.append(pinfo)
        if len(pinfo) == 0:
            return None
        else:
            return found

    @classmethod
    def bash(cls, *args):
        """
        executes bash with the given arguments
        :param args:
        :return:
        """
        return cls.execute('bash', args)

    @classmethod
    # @NotImplementedInWindows
    def brew(cls, *args):
        """
        executes bash with the given arguments
        :param args:
        :return:
        """
        NotImplementedInWindows()
        return cls.execute('brew', args)

    @classmethod
    # @NotImplementedInWindows
    def cat(cls, *args):
        """
        executes cat with the given arguments
        :param args:
        :return:
        """
        if os_is_windows() and is_gitbash():
            content = readfile(args[0])
            return content
        else:
            return cls.execute('cat', args)

    @classmethod
    def git(cls, *args):
        """
        executes git with the given arguments
        :param args:
        :return:
        """
        return cls.execute('git', args)

    # noinspection PyPep8Naming
    @classmethod
    def VBoxManage(cls, *args):
        """
        executes VboxManage with the given arguments
        :param args:
        :return:
        """

        if platform == "darwin":
            command = "/Applications/VirtualBox.app/Contents/MacOS/VBoxManage"
        else:
            command = 'VBoxManage'
        return cls.execute(command, args)

    @classmethod
    def blockdiag(cls, *args):
        """
        executes blockdiag with the given arguments
        :param args:
        :return:
        """
        return cls.execute('blockdiag', args)

    @classmethod
    def cm(cls, *args):
        """
        executes cm with the given arguments
        :param args:
        :return:
        """
        return cls.execute('cm', args)

    @classmethod
    def cms(cls, *args):
        """
        executes cm with the given arguments
        :param args:
        :return:
        """
        return cls.execute('cms', args)

    @classmethod
    def cmsd(cls, *args):
        """
        executes cm with the given arguments
        :param args:
        :return:
        """
        return cls.execute('cmsd', args)

    @classmethod
    def head(cls, filename=None, lines=10):
        """
        executes head with the given arguments
        :param args:
        :return:
        """
        filename = cls.map_filename(filename).path
        r = Shell.run(f'head -n {lines} {filename}')
        return r

    @classmethod
    def keystone(cls, *args):
        """
        executes keystone with the given arguments
        :param args:
        :return:
        """
        return cls.execute('keystone', args)

    @staticmethod
    def kill_pid(pid):

        if sys.platform == 'win32':
            try:
                result = Shell.run(f"taskkill /IM {pid} /F")
            except Exception as e:
                result = str(e)
        else:
            try:
                result = Shell.kill("-9", pid)
            except subprocess.CalledProcessError:
                result = 'server is already down...'
        return result

    @classmethod
    # @NotImplementedInWindows
    def kill(cls, *args):
        """
        executes kill with the given arguments
        :param args:
        :return:
        """
        NotImplementedInWindows()
        # TODO: use tasklisk, compare to linux
        return cls.execute('kill', args)

    @classmethod
    def download(cls, source, destination, force=False, provider=None, chunk_size=128):
        """
        Given a source url and a destination filename, download the file at the source url
        to the destination.

        If provider is None, the request lib is used
        If provider is 'system', wget, curl, and requests lib are attempted in that order
        """
        destination = path_expand(destination)

        if os.path.exists(destination) and not force:
            return destination

        if provider == 'system':
            # First try wget
            wget_return = os.system(f'wget -O {destination} {source}')
            if wget_return == 0:
                Console.info('Used wget')
                return destination
            # Then curl
            curl_return = os.system(f'curl -L -o {destination} {source}')
            if curl_return == 0:
                Console.info('Used curl')
                return destination
        # Default is requests lib. If provider is None, or if provider == 'system'
        # but wget and curl fail, default here
        r = requests.get(source, stream=True, allow_redirects=True)
        total_size = int(r.headers.get('content-length'))

        with open(destination, 'wb') as fd:
            with tqdm(total=total_size, unit="B",
                      unit_scale=True, desc=destination, initial=0, ascii=True) as pbar:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    fd.write(chunk)
                    pbar.update(len(chunk))

            return destination

    @classmethod
    def mount(cls, *args):
        """
        mounts a given mountpoint to a file
        :param args:
        :return:
        """
        return cls.execute('mount', args)

    @classmethod
    def umount(cls, *args):
        """
        umounts a given mountpoint to a file
        :param args:
        :return:
        """
        return cls.execute('umount', args)

    @classmethod
    def nova(cls, *args):
        """
        executes nova with the given arguments
        :param args:
        :return:
        """
        return cls.execute('nova', args)

    @classmethod
    def ping(cls, host=None, count=1):
        """
        execute ping
        :param host: the host to ping
        :param count: the number of pings
        :return:
        """
        r = None
        option = '-n' if os_is_windows() else '-c'
        parameters = "{option} {count} {host}".format(option=option,
                                                      count=count,
                                                      host=host)
        r = Shell.run(f'ping {parameters}')
        if r is None:
            Console.error("ping is not installed")
        return r

    @classmethod
    def pwd(cls, *args):
        """
        executes pwd with the given arguments
        :param args:
        :return:
        """
        return os.getcwd()

    @classmethod
    def rackdiag(cls, *args):
        """
        executes rackdiag with the given arguments
        :param args:
        :return:
        """
        return cls.execute('rackdiag', args)

    @classmethod
    def rm(cls, location):
        """
        executes rm with the given arguments
        :param args:
        :return:
        """
        location = cls.map_filename(location).path
        os.remove(location)

    @classmethod
    def rsync(cls, *args):
        """
        executes rsync with the given arguments
        :param args:
        :return:
        """
        return cls.execute('rsync', args)

    @classmethod
    def scp(cls, *args):
        """
        executes scp with the given arguments
        :param args:
        :return:
        """
        return cls.execute('scp', args)

    @classmethod
    # @NotImplementedInWindows
    def sort(cls, *args):
        """
        executes sort with the given arguments
        :param args:
        :return:
        """
        NotImplementedInWindows()
        # TODO: https://superuser.com/questions/1316317/is-there-a-windows-equivalent-to-the-unix-uniq
        return cls.execute('sort', args)

    @classmethod
    def sh(cls, *args):
        """
        executes sh with the given arguments
        :param args:
        :return:
        """
        return cls.execute('sh', args)

    @classmethod
    def ssh(cls, *args):
        """
        executes ssh with the given arguments
        :param args:
        :return:
        """
        return cls.execute('ssh', args)

    @classmethod
    # @NotImplementedInWindows
    def sudo(cls, *args):
        """
        executes sudo with the given arguments
        :param args:
        :return:
        """
        NotImplementedInWindows()
        # TODO: https://stackoverflow.com/questions/9652720/how-to-run-sudo-command-in-windows
        return cls.execute('sudo', args)

    @classmethod
    def tail(cls, filename=None, lines=10):
        """
        executes tail with the given arguments
        :param args:
        :return:
        """
        filename = cls.map_filename(filename).path
        r = Shell.run(f'tail -n {lines} {filename}')
        return r

    @classmethod
    def vagrant(cls, *args):
        """
        executes vagrant with the given arguments
        :param args:
        :return:
        """
        return cls.execute('vagrant', args, shell=True)

    @classmethod
    def pandoc(cls, *args):
        """
        executes vagrant with the given arguments
        :param args:
        :return:
        """
        return cls.execute('pandoc', args)

    @classmethod
    def mongod(cls, *args):
        """
        executes mongod with the given arguments
        :param args:
        :return:
        """
        return cls.execute('mongod', args)

    @classmethod
    # @NotImplementedInWindows
    def dialog(cls, *args):
        """
        executes dialof with the given arguments
        :param args:
        :return:
        """
        NotImplementedInWindows()
        return cls.execute('dialog', args)

    @classmethod
    def pip(cls, *args):
        """
        executes pip with the given arguments
        :param args:
        :return:
        """
        return cls.execute('pip', args)

    @classmethod
    def fgrep(cls, string=None, file=None):
        if not os_is_windows():
            r = Shell.run(f'fgrep {string} {file}')
        else:
            r = Shell.run(f'grep -F {string} {file}')
        return r

    @classmethod
    def grep(cls, string=None, file=None):
        r = Shell.run(f'grep {string} {file}')
        return r

    @classmethod
    def cm_grep(cls, lines, what):
        """
        returns all lines that contain what
        :param lines:
        :param what:
        :return:
        """
        if type(lines) == str:
            _lines = lines.splitlines()
        else:
            _lines = lines
        result = []
        for line in _lines:
            if what in line:
                result.append(line)
        return result

    @classmethod
    def remove_line_with(cls, lines, what):
        """
        returns all lines that do not contain what
        :param lines:
        :param what:
        :return:
        """
        if type(lines) == str:
            _lines = lines.splitlines()
        else:
            _lines = lines
        result = []
        for line in _lines:
            if what not in line:
                result = result + [line]
        return result

    @classmethod
    def find_lines_with(cls, lines, what):
        """
        returns all lines that contain what
        :param lines:
        :param what:
        :return:
        """
        if type(lines) == str:
            _lines = lines.splitlines()
        else:
            _lines = lines
        result = []
        for line in _lines:
            if what in line:
                result = result + [line]
        return result

    @classmethod
    def find_lines_from(cls, lines, what):
        """
        returns all lines that come after a particular line
        :param lines:
        :param what:
        :return:
        """
        if type(lines) == str:
            _lines = lines.splitlines()
        else:
            _lines = lines
        result = []
        found = False
        for line in _lines:
            found = found or what in line
            if found:
                result = result + [line]
        return result

    @classmethod
    def find_lines_between(cls, lines, what_from, what_to):
        """
        returns all lines that come between the markers
        :param lines:
        :param what:
        :return:
        """
        select = Shell.find_lines_from(lines, what_from)
        select = Shell.find_lines_to(select, what_to)
        return select

    @classmethod
    def find_lines_to(cls, lines, what):
        """
        returns all lines that come before a particular line
        :param lines:
        :param what:
        :return:
        """
        if type(lines) == str:
            _lines = lines.splitlines()
        else:
            _lines = lines
        result = []
        found = True
        for line in _lines:
            if what in line:
                return result
            else:
                result = result + [line]
        return result

    @classmethod
    def terminal_type(cls):
        """
        returns  darwin, cygwin, cmd, or linux
        """
        what = sys.platform

        kind = 'UNDEFINED_TERMINAL_TYPE'
        if 'linux' in what:
            kind = 'linux'
        elif 'darwin' in what:
            kind = 'darwin'
        elif 'cygwin' in what:
            kind = 'cygwin'
        elif what in ['windows', 'win32']:
            kind = 'windows'

        return kind

    @classmethod
    def which(cls, command):
        """
        returns the path of the command with which
        :param command: teh command
        :return: the path
        """
        if os_is_windows():
            return Shell.run(f"where {command}")
        else:
            return shutil.which(command)

    @classmethod
    def command_exists(cls, name):
        """
        returns True if the command exists
        :param name:
        :return:
        """
        return cls.which(name) is not None

    @classmethod
    def operating_system(cls):
        """
        the name of the os
        :return: the name of the os
        """
        return platform

    @staticmethod
    def get_pid(name, service="psutil"):
        pid = None
        for proc in psutil.process_iter():
            if name in proc.name():
                pid = proc.pid
        return pid

    @staticmethod
    def cms(command, encoding='utf-8'):
        return Shell.run("cms " + command, encoding=encoding)

    @classmethod
    def check_python(cls):
        """
        checks if the python version is supported
        :return: True if it is supported
        """
        python_version = sys.version_info[:3]

        v_string = [str(i) for i in python_version]

        python_version_s = '.'.join(v_string)

        if python_version[0] == 2:

            print(
                "You are running an unsupported version of python: {:}".format(
                    python_version_s))

            # python_version_s = '.'.join(v_string)
            # if (python_version[0] == 2) and (python_version[1] >= 7) and (python_version[2] >= 9):

            #    print("You are running an unsupported version of python: {:}".format(python_version_s))
            # else:
            #    print("WARNING: You are running an unsupported version of python: {:}".format(python_version_s))
            #    print("         We recommend you update your python")

        elif python_version[0] == 3:

            if (python_version[0] == 3) and (python_version[1] >= 7) and (python_version[2] >= 0):

                print(
                    "You are running a supported version of python: {:}".format(
                        python_version_s))
            else:
                print(
                    "WARNING: You are running an unsupported version of python: {:}".format(
                        python_version_s))
                print("         We recommend you update your python")

        # pip_version = pip.__version__
        python_version, pip_version = cls.get_python()

        if int(pip_version.split(".")[0]) >= 19:
            print("You are running a supported version of pip: " + str(
                pip_version))
        else:
            print("WARNING: You are running an old version of pip: " + str(
                pip_version))
            print("         We recommend you update your pip  with \n")
            print("             pip install -U pip\n")

    @classmethod
    def copy(cls, source, destination, expand=False):
        if expand:
            s = path_expand(source)
            d = path_expand(destination)
        else:
            s = source
            d = destination
        shutil.copy2(s, d)

    @classmethod
    def mkdir(cls, directory):
        """
        creates a directory with all its parents in ots name
        :param directory: the path of the directory
        :return:
        """
        directory = cls.map_filename(directory).path
        try:
            os.makedirs(directory)
            return True
        except OSError as e:
            return False

    def unzip(cls, source_filename, dest_dir):
        """
        unzips a file into the destination directory
        :param source_filename: the source
        :param dest_dir: the destination directory
        :return:
        """

        with zipfile.ZipFile(source_filename) as zf:
            for member in zf.infolist():
                # Path traversal defense copied from
                # http://hg.python.org/cpython/file/tip/Lib/http/server.py#l789
                words = member.filename.split('/')
                path = path_expand(dest_dir)
                for word in words[:-1]:
                    drive, word = os.path.splitdrive(word)
                    head, word = os.path.split(word)
                    if word in (os.curdir, os.pardir, ''):
                        continue
                    path = os.path.join(path, word)
                zf.extract(member, path)

    @staticmethod
    def edit(filename):
        if platform == 'darwin':
            os.system("emacs " + filename)
        elif platform == "windows":
            os.system("notepad " + filename)
        else:
            raise NotImplementedError("Editor not configured for OS")

    @classmethod
    # @NotImplementedInWindows
    def lsb_release(cls):
        """
        executes lsb_release command
        :param args:
        :return:
        """
        NotImplementedInWindows()
        return cls.execute('lsb_release', ['-a'])

    @classmethod
    def distribution(cls):
        """
        executes lsb_release command
        :param args:
        :return:

        TODO: needs testing
        """

        machine = platform.lower()

        result = {"platform": machine,
                  "distribution": None}

        if machine == "linux":
            try:
                release = readfile("/etc/os-release")
                for line in release.splitlines():
                    attribute, value = line.split("=", 1)
                    result[attribute] = value
                if "Debian" in result["NAME"]:
                    result["distribution"] = "debian"
                elif "Ubuntu" in result["NAME"]:
                    result["distribution"] = "ubuntu"

            except:  # noqa: E722
                try:
                    r = cls.lsb_release()
                    for line in r.split():
                        if ":" in line:
                            attribute, value = line.split(":", 1)
                            attribute = attribute.strip().replace(" ",
                                                                  "_").lower()
                            value = value.strip()
                            result[attribute] = value
                    result["distribution"] = result["description"].split(" ")[
                        0].lower()
                except:  # noqa: E722
                    Console.error(
                        f"lsb_release not found for the platform {machine}")
                    raise NotImplementedError
        elif machine == 'darwin':
            result["distribution"] = "macos"
            result["version"] = os_platform.mac_ver()[0]
        elif machine == 'win32':
            result["distribution"] = "windows"
            result["version"] = os_platform.win_ver()[0]
        else:
            Console.error(f"not implemented for the platform {machine}")
            raise NotImplementedError

        return result

    @staticmethod
    def open(filename=None, program=None):
        if not os.path.isabs(filename):
            filename = path_expand(filename)

        if os_is_linux():
            r = os.system(f"gopen {filename}")
        if os_is_mac():
            command = f'open {filename}'
            if program:
                command += f' -a "{program}"'
            r = Shell.run(command)
        if os_is_windows():
            r = Shell.run(f"start {filename}")

        return r

    @staticmethod
    def sys_user():
        if os_is_windows():
            localuser = os.environ["USERNAME"]
        else:
            try:
                localuser = os.environ['USER']
            except:
                # docker image does not have user variable. so just do whoami
                localuser = Shell.run('whoami')
        return localuser


def main():
    """
    a test that should actually be added into a pytest
    :return:
    """

    print(Shell.terminal_type())

    r = Shell.execute('pwd')  # copy line replace
    print(r)

    # shell.list()

    # print json.dumps(shell.command, indent=4)

    # test some commands without args
    """
    for cmd in ['whoami', 'pwd']:
        r = shell._execute(cmd)
        print ("---------------------")
        print ("Command: {:}".format(cmd))
        print ("{:}".format(r))
        print ("---------------------")
    """
    r = Shell.execute('ls', ["-l", "-a"])
    print(r)

    r = Shell.execute('ls', "-l -a")
    print(r)

    if sys.platform != "win32":
        r = Shell.unix_ls("-aux")
        print(r)

        r = Shell.unix_ls("-a", "-u", "-x")
        print(r)

    r = Shell.ls("./*.py")
    print(r)

    r = Shell.ls("*/*.py")
    print(r)

    r = Shell.pwd()
    print(r)


if __name__ == "__main__":
    main()



# noinspection PyPep8Naming
class dotdict(dict):
    """
    A convenient dot dict class::

        a = dotdict({"argument": "value"})

    print (a.argument)

    Nested dot documentation is not supported.
    """

    def __getattr__(self, attr):
        """
        retirns an element

        :param attr: the attribute key
        :return: teh value
        """
        return self.get(attr)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



import os
import subprocess
from cloudmesh.common.util import banner


class Sudo:

    @staticmethod
    def password(msg="sudo password: "):
        """
        Asks for the Sudo password
        """
        os.system(f'sudo -p "{msg}"  echo "" > /dev/null')

    @staticmethod
    def expire():
        """
        Expires the password
        """
        os.system("sudo -k")

    @staticmethod
    def execute(command, decode="True", debug=False, msg=None):
        """
        Executes the command

        :param command: The command to run
        :type command: list or str
        :return:
        :rtype:
        """

        Sudo.password()
        if type(command) == str:
            sudo_command = "sudo " + command
            sudo_command = sudo_command.split(" ")
        else:
            sudo_command = ["sudo"] + command
        if msg is None:
            pass
        elif msg == "command":
            print("Executing:", " ".join(sudo_command))
        else:
            print("Executing:", " ".join(msg))
        os.system("sync")
        result = subprocess.run(sudo_command, capture_output=True)
        os.system("sync")
        if decode:
            result.stdout = result.stdout.decode('ascii')
            result.stderr = result.stderr.decode('ascii')

        if debug:
            banner("stdout")
            print(result.stdout)
            banner("stderr")
            print(result.stderr)
            banner("result")
            print(result)
        return result

    @staticmethod
    def readfile(filename, split=False, trim=False, decode=True):
        """
        Reads the content of the file as sudo and returns the result

        :param filename: the filename
        :type filename: str
        :param split: uf true returns a list of lines
        :type split: bool
        :param trim: trim trailing whitespace. This is useful to
                     prevent empty string entries when splitting by '\n'
        :type trim: bool
        :return: the content
        :rtype: str or list
        """
        Sudo.password()
        os.system("sync")
        result = Sudo.execute(f"cat {filename}", decode=decode)

        content = result.stdout

        if trim:
            content = content.rstrip()

        if split:
            content = content.splitlines()

        return content

    @staticmethod
    def writefile(filename, content, append=False):
        """
        Writes the content in the the given file.

        :param filename: the filename
        :type filename: str
        :param content: the content
        :type content: str
        :param append: if true it append it at the end, otherwise the file will
                       be overwritten
        :type append: bool
        :return: the output created by the write process
        :rtype: int
        """

        os.system("sync")
        Sudo.password()
        if append:
            content = Sudo.readfile(filename, split=False, decode=True) + content

        os.system(f"echo '{content}' | sudo cp /dev/stdin {filename}")
        os.system("sync")

        return content



"""
cloudmesh name space.
"""
# noa: E402
import pkg_resources

pkg_resources.declare_namespace(__name__)

import os
import shutil
from cloudmesh.common.util import path_expand
from cloudmesh.common.console import Console
from cloudmesh.common.Shell import Shell

import sys
IN_COLAB = 'google.colab' in sys.modules

if IN_COLAB:
    os.environ["USER"] = "COLAB"



import os
import platform
import subprocess
from collections import OrderedDict
from multiprocessing import Pool
from pprint import pprint

from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.dotdict import dotdict
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile


# TODO: pytest
class JobSet:
    """
    JobSet is a general execution framework for running a set of jobs on which
    we specify a self defined job executor function. Through this framework it
    is possible to very flexibly integrate different JobSets on which are
    executed based on the executor. The jobset is executed in parallel and a run
    method can be augmented with the number of parallel jobs being executed at
    the same time. A simple executor that runs commands in the shell is provided

    Basic Example Usage:

        t = JobSet("terminal-commands", executor=JobSet.execute)
        t.add({"name": "pwd", "command": "pwd"})
        t.add({"name": "info", "command": "uname -a"})
        t.add({"name": "uname", "command": "uname"})
        t.add({"name": "hostname", "command": "hostname"})
        t.add({"name": "provccesses", "command": "ps"})
        t.run(parallel=3)
        t.Print()

    Advanced Example Usage:

        from cloudmesh.common.parameter import Parameter

        def ssh(spec):
             result = subprocess.check_output("ssh " + spec["command"], shell=True)
             success = "Error" noyt in result
             if success:
                returncode = 0
                status = 'done'
            else:
                returncode = 1
                status = 'failed'

            return dict({
                "name": spec["name"],
                "stdout": result,
                "stderr": ""
                "returncode": returncode
                "status": status
            })

        t = JobSet("terminal-commands", executor=JobSet.execute)
        for host in Parameter.expand("red[01-03]"):
            t.add({"name": "host", "command": "uname -a"})
        t.run(parallel=3)
        t.Print()

    """

    def __init__(self, name, executor=None):
        self.name = name
        self.job = OrderedDict({})
        self.executor = executor or JobSet.execute

    def reset(self, name, executor=None):
        self.name = name
        self.job = OrderedDict({})
        self.executor = executor or JobSet.execute

    @staticmethod
    def ssh(spec):
        """

        name: name of the job
        host: host on which we execute
        os:   if true use os.system, this uses a temporary file, so be careful
              if false use subprocess.check_output
        stderr: result of stderror
        stdout: result of stdout
        returncode: 0 is success
        success: Ture if successfull e.g. returncode == 0
        status: a status: defined, running, done, failed

        :param spec:
        :return:
        """
        spec = dotdict(spec)
        hostname = platform.uname()[1]
        local = hostname == spec.host

        if 'key' not in spec:
            spec.key = path_expand("~/.ssh/id_rsa.pub")

        ssh = \
            f'ssh' \
            f' -o StrictHostKeyChecking=no' \
            f' -o UserKnownHostsFile=/dev/null' \
            f' -i {spec.key} {spec.host}'

        # result.stdout = result.stdout.decode("utf-8", "ignore")

        if "os" in spec:

            if 'tmp' not in spec:
                spec.tmp = "/tmp"
            tee = f"tee {spec.tmp}/cloudmesh.{spec.name}"
            if local:
                command = f"{spec.command} | {tee}"
            else:
                command = f"{ssh} {spec.command} | {tee}"
            # print ("RUN os.system", command)

            returncode = os.system(command)
            result = readfile(f"{spec.tmp}/cloudmesh.{spec.name}").strip()
        else:
            if local:
                command = f"{spec.command} "
            else:
                command = f"{ssh} \'{spec.command}\'"
            # print ("RUN check_output", command)

            try:
                stderr = ""
                returncode = 0
                result = subprocess.check_output(command, shell=True, stderr=subprocess.PIPE)
            except subprocess.CalledProcessError as grepexc:
                # print("Error Code", grepexc.returncode, grepexc.output)
                result = "Command could not run | Error Code: ", grepexc.returncode
                # result = grepexc.stdout
                stderr = grepexc.stderr
                returncode = grepexc.returncode

        return dict({
            "name": spec.name,
            "stdout": result,
            "stderr": stderr,
            "returncode": returncode,
            "status": "defined"
        })

    @staticmethod
    def execute(spec):
        result = subprocess.check_output(spec["command"], shell=True)

        return dict({
            "name": spec["name"],
            "stdout": result,
            "stderr": "",
            "returncode": 0,
            "status": "defined"
        })

    @staticmethod
    def identity(entry_with_name):

        return dict({
            "name": entry_with_name["name"],
            "stdout": "identity " + entry_with_name["name"],
            "stderr": "",
            "returncode": 0,
            "status": "defined"
        })

    def add(self, spec, executor=None):
        name = spec["name"]
        self.job[name] = spec
        self.job[name]["status"] = "defined"
        self.job[name]["executor"] = spec.get(
            "executor") or executor or self.executor

    def _run(self, spec):
        result = dict(spec)
        result["status"] = "running"
        executor = spec["executor"]
        res = executor(spec)
        result.update(res)
        result["status"] = "done"
        return result

    def run(self, parallel=3):

        if len(self.job) == 0:
            res = None
        elif len(self.job) == 1:
            id = next(iter(self.job))
            job = self.job[id]
            res = self._run(job)
            self.job[id].update(res)
        else:
            joblist = [self.job[x] for x in self.job]
            # VERBOSE(joblist)
            with Pool(parallel) as p:
                res = p.map(self._run, joblist)
                p.close()
                p.join()

            for entry in res:
                name = entry['name']
                for a in entry:
                    self.job[name].update(entry)

        return res

    def __len__(self):
        return len(self.job)

    def __repr__(self):
        return str(self.job)

    def __str__(self):
        return str(self.job)

    def Print(self):
        print()
        d = dict(self.job)
        for e in d:
            del d[e]["executor"]
        pprint(d)

    def array(self):
        return [self.job[x] for x in self.job]


if __name__ == '__main__':
    def command_execute(spec):
        return dict({
            "name": spec["name"],
            "stdout": "command " + spec["command"],
            "stderr": "",
            "returncode": 0,
            "status": "defined"
        })


    hostname = platform.uname()[1]

    s = JobSet("test", executor=JobSet.identity)

    s.add({"name": "a", "value": 1})
    s.add({"name": "b", "value": 2})

    pprint(s.job)
    result = s.run()
    pprint(result)

    pprint(s.job)
    s.Print()

    t = JobSet("test2", executor=JobSet.execute)
    t.add({"name": "c", "command": "pwd"})
    t.add({"name": "d", "command": "uname -a"})
    t.add({"name": "e", "command": "uname"})
    t.add({"name": "f", "command": "hostname"})
    t.add({"name": "g", "command": "ps"})
    t.run()
    t.Print()

    # x = {"name": "x", "command": "ls", "executor": command_execute}
    # print (command_execute(x))

    t = JobSet("onejob", executor=JobSet.execute)
    t.add({"name": "c", "command": "pwd"})
    t.run()
    t.Print()

    t = JobSet("ssh", executor=JobSet.ssh)
    t.add({"name": hostname, "host": hostname, "command": "pwd"})
    t.run()
    t.Print()

    t = JobSet("onejob", executor=JobSet.ssh)
    t.add({"name": hostname, "host": hostname, "command": "pwd", "os": True})
    t.run()
    t.Print()

    t = JobSet("onejob", executor=JobSet.ssh)
    t.add({"name": hostname, "host": hostname, "command": "pwd", "os": False})
    t.run()
    t.Print()

    t = JobSet("onejob", executor=JobSet.ssh)
    t.add({"name": hostname, "host": "red", "command": "pwd", "os": False})
    t.run()
    t.Print()

    hostname = platform.uname()[1]
    t = JobSet("onejob", executor=JobSet.ssh)
    for host in Parameter.expand("red,red[01-03]"):
        t.add({"name": host, "host": host, "command": "uname -a"})
    t.add({"name": hostname, "host": hostname, "command": "uname -a"})

    t.run(parallel=3)
    print(Printer.write(t.array(),
                        order=["name", "command", "status", "stdout",
                               "returncode"]))



from cloudmesh.common.JobSet import JobSet
import os
import platform
import textwrap
from cloudmesh.common.dotdict import dotdict
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.console import Console


class JobScript:
    '''
    The jobscript is a simple mechanism to run a number of commands specified
    in a script. The script is interpreted line by line and does not support
    multi line commands at this time. (Not difficult to implement when
    looking at \\ at the end of a line.)

    It can be either used a static call or as variable invocation. Often the
    static call will be sufficient.

    Static method::

        from cloudmesh.common.JobSet import JobScript
        from cloudmesh.common.Tabulate import Printer

        result = JobScript.execute("""
            # This is a comment

            pwd                             # tag: pwd
            uname -a
        """)
        print(Printer.write(result,
                order=["name", "command", "status", "stdout", "returncode"]))

    Variable invocation::

        from cloudmesh.common.Tabulate import Printer

        job = JobScript()
        job.run(name="variable script",
                script="""
                    # This is a comment

                    pwd                    # tag: pwd
                    uname -a
                """)
        print(Printer.write(
            job.array(),
            order=["name", "command", "status", "stdout", "returncode"]))

        Each line is augmented with a name, so you can easily retrive the result
        content by that name. By default the name is the line number, however it
        can be overwritten with `# tag:` at the end of the line. The line number
        starts at 1.

    Partial example output:

        # +------+-----------+--------+------------------------+--------------+
        # | name | command   | status |                        |   returncode |
        # |------+-----------+--------+------------------------+--------------|
        # | pwd  | pwd       | done   | /Users/.../cloudmesh   |            0 |
        # | 4    | uname -a  | done   | Darwin gray 19.4.0 ... |            0 |
        # +------+-----------+--------+------------------------+--------------+
    '''

    def __init__(self):
        self.script = None

    def run(self, script=None, name="script", host=None, executor=JobSet.ssh,
            **kwargs):
        # Prepare parameters
        if script is None:
            Console.error("The script is not defined, found None as content")
            return
        host = host or platform.uname()[1]
        if kwargs:
            parameters = dotdict(**kwargs)
        else:
            parameters = dotdict({})
        parameters.host = host
        # Prepare script
        self.script = textwrap.dedent(str(script))
        self.script = self.script.format(**kwargs)

        lines = self.script.splitlines()
        # Add script to jobset and run
        jobs = JobSet("onejob", executor=executor)
        counter = 1
        for line in lines:
            stripped = line.strip()
            if stripped.startswith("#") or stripped == "":
                pass
            else:
                tag = counter
                if "# tag:" in line:
                    line, tag = line.split("# tag:", 1)
                    tag = tag.strip()
                    line = line.strip()
                jobs.add_directory({
                    "script": name,
                    "name": tag,
                    "tag": tag,
                    "line": counter,
                    "host": host,
                    "counter": counter,
                    "command": line
                })
            counter = counter + 1
        jobs.run(parallel=1)
        self.job = jobs.job
        return self.job

    @staticmethod
    def execute(script, name="script", host=None, executor=JobSet.ssh,
                **kwargs):
        job = JobScript()
        job.run(script=script, name=name, host=host, **kwargs)
        return job.array()

    @staticmethod
    def _array():
        return

    def array(self):
        return [self.job[x] for x in self.job]


if __name__ == '__main__':
    #
    # Tow ways to run.
    #

    # Static method
    result = JobScript.execute("""
        # This is a comment
        pwd                             # tag: pwd
        uname -a
    """)
    print(Printer.write(
        result,
        order=["name", "command", "status", "stdout", "returncode"]))

    # Variables
    job = JobScript()
    job.run(name="variable script",
            script="""
                # This is a comment
                pwd                    # tag: pwd
                uname -a
            """)
    print(Printer.write(
        job.array(),
        order=["name", "command", "status", "stdout", "returncode"]))

#
# Partial example output
#
# +------+-----------+--------+-------------------------------+--------------+
# | name | command   | status |                               |   returncode |
# |------+-----------+--------+-------------------------------+--------------|
# | pwd  | pwd       | done   | /Users/..../cloudmesh/common  |            0 |
# | 4    | uname -a  | done   | Darwin gray 19.4.0 Darwin ... |            0 |
# +------+-----------+--------+-------------------------------+--------------+



import os
import platform as platform_module
import subprocess
import textwrap
from multiprocessing import Pool
from sys import platform
from time import sleep
from random import randint

from cloudmesh.common.DateTime import DateTime
from cloudmesh.common.parameter import Parameter
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import readfile
from pprint import pprint
from cloudmesh.common.Printer import Printer


class Host(object):

    def _print(results, output='table'):

        if output in ['table', 'yaml']:
            print(Printer.write(results,
                                order=['host', 'success', 'stdout', 'stderr'],
                                output=output))
        else:
            pprint(results)

    @staticmethod
    def get_hostnames(names):
        """
        Given a list of host names it identifies if they have numbers in them. If so, they are assumed workers.
        If not, it is a manager. There can only be one manager.

        @param names: list of names
        @type names: str
        @return: manager, worker as list
        @rtype: tuple
        """
        manager = None
        workers = []
        for name in names:
            if any(map(str.isdigit, name)):
                workers.append(name)
            else:
                manager = name

        if len(workers) == 0:
            workers = None

        return manager, workers

    @staticmethod
    def config(hosts=None,
               ips=None,
               username=None,
               key="~/.ssh/id_rsa"):

        if type(hosts) != list:
            _hosts = Parameter.expand(hosts)
        if ips is not None:
            if type(ips) != list:
                _ips = Parameter.expand(ips)
            if len(_ips) != len(_hosts):
                raise ValueError("Number of hosts and ips mismatch")

        result = ""
        for i in range(0, len(_hosts)):

            host = _hosts[i]

            hostname = ""
            user = f"user {username}"
            if ips is None:
                hostname = ""
            else:
                ip = _ips[i]
                hostname = f"Hostname {ip}"

            data = textwrap.dedent(f"""
            Host {host}
                StrictHostKeyChecking no
                LogLevel ERROR
                UserKnownHostsFile /dev/null
                IdentityFile {key}
                {hostname}
            """)

            if username:
                data += (f"    {user}\n")

            result += data
        return result

    @staticmethod
    def _run(args):
        """

        An internal command that executes as part of a process map a given
        command. args is a dict and must include

        * command
        * shell

        It returns a dict of the form

        * command
        * stdout
        & stderr
        * returncode
        * success

        :param args: command dict
        :return:
        """
        try:
            # experimental sleep as we get a block on ssh commands

            hostname = platform_module.uname()[1]
            host = args.get("host")

            shell = args.get("shell")

            if host == hostname:
                command = args.get("execute")
                result = subprocess.getoutput(command)
                stderr = ""
                returncode = 0
                stdout = result
            else:
                command = args.get("command")

                result = subprocess.run(
                    command,
                    capture_output=True,
                    shell=shell)

                result.stdout = result.stdout.decode("utf-8", "ignore").strip()
                if result.stderr == b'':
                    result.stderr = None

                stderr = result.stderr
                returncode = result.returncode
                stdout = result.stdout

            data = {
                'host': args.get("host"),
                'command': args.get("command"),
                'execute': args.get("execute"),
                'stdout': stdout,
                'stderr': stderr,
                'returncode': returncode,
                'success': returncode == 0,
                'date': DateTime.now(),
                'cmd': " ".join(args.get("command"))
            }
        except Exception as e:
            print(e)
            data = None
        return data

    @staticmethod
    def run(hosts=None,
            command=None,
            execute=None,
            processors=3,
            shell=False,
            **kwargs):
        """
        Executes the command on all hosts. The key values
        specified in **kwargs will be replaced prior to the
        execution. Furthermore, {host} will be replaced with the
        specific hostname.

        :param hosts: The hosts given in parameter notation
                      Example: red[01-10]
        :param command: The command to be executed for each host
                        Example: ssh {host} uname
        :param username: Specify the username on the host
        :param processors: The number of parallel processes used
        :param shell: Set to Tue if the current context of the shell is
                      to be used. It is by default True
        :param kwargs: The key value pairs to be replaced in the command
        :return:
        """

        hosts = Parameter.expand(hosts)
        args = [{'command': [c.format(host=host, **kwargs) for c in command],
                 'shell': shell,
                 'host': host,
                 'execute': execute,
                 } for host in hosts]

        if "executor" not in args:
            _executor = Host._run
        # from pprint import pprint
        # os.sync()
        # pprint(args)
        with Pool(processors) as p:
            res = p.map(_executor, args)
            p.close()
            p.join()
        return res

    @staticmethod
    def ssh(hosts=None,
            command=None,
            username=None,
            key="~/.ssh/id_rsa",
            processors=3,
            dryrun=False,  # notused
            executor=None,
            verbose=False,  # not used
            **kwargs):
        #
        # BUG: this code has a bug and does not deal with different
        #  usernames on the host to be checked.
        #
        """

        :param command: the command to be executed
        :param hosts: a list of hosts to be checked
        :param username: the usernames for the hosts
        :param key: the key for logging in
        :param processors: the number of parallel checks
        :return: list of dicts representing the ping result
        """

        hosts = Parameter.expand(hosts)

        key = path_expand(key)

        ssh_command = ['ssh',
                       '-o', 'StrictHostKeyChecking=no',
                       '-o', 'UserKnownHostsFile=/dev/null',
                       '-i', key,
                       '{host}',
                       f'{command}']
        result = Host.run(hosts=hosts,
                          command=ssh_command,
                          execute=command,
                          shell=False,
                          executor=executor,
                          **kwargs)

        return result

    @staticmethod
    def put(hosts=None,
            source=None,
            destination=None,
            username=None,
            key="~/.ssh/id_rsa",
            shell=False,
            processors=3,
            dryrun=False,
            verbose=False):
        """

        :param command: the command to be executed
        :param hosts: a list of hosts to be checked
        :param username: the usernames for the hosts
        :param key: the key for logging in
        :param processors: the number of parallel checks
        :return: list of dicts representing the ping result
        """

        hosts = Parameter.expand(hosts)

        key = path_expand(key)

        command = ['scp',
                   "-o", "StrictHostKeyChecking=no",
                   "-o", "UserKnownHostsFile=/dev/null",
                   '-i', key,
                   source,
                   "{host}:{destination}"]

        execute = f"cp {source} {destination}"

        result = Host.run(hosts=hosts,
                          command=command,
                          execute=execute,
                          destination=destination,
                          shell=False)

        return result

    @staticmethod
    def check(hosts=None,
              username=None,
              key="~/.ssh/id_rsa",
              processors=3):
        #
        # BUG: this code has a bug and does not deal with different
        #  usernames on the host to be checked.
        #
        """

        :param hosts: a list of hosts to be checked
        :param username: the usernames for the hosts
        :param key: the key for logging in
        :param processors: the number of parallel checks
        :return: list of dicts representing the ping result
        """
        hosts = Parameter.expand(hosts)

        result = Host.ssh(hosts=hosts,
                          command='hostname',
                          username=username,
                          key=key,
                          processors=processors)

        return result

    # noinspection PyBroadException,PyPep8
    @staticmethod
    def _ping(args):
        """
            ping a vm

            :param args: dict of {ip address, count}
            :return: a dict representing the result, if returncode=0 ping is
                     successfully
            """
        ip = args['ip']
        count = str(args['count'])
        count_flag = '-n' if platform == 'windows' else '-c'
        command = ['ping', count_flag, count, ip]
        result = subprocess.run(command, capture_output=True)

        try:
            timers = result.stdout \
                .decode("utf-8", "ignore") \
                .split("round-trip min/avg/max/stddev =")[1] \
                .replace('ms', '').strip() \
                .split("/")
            data = {
                "host": ip,
                "success": result.returncode == 0,
                "stdout": result.stdout,
                "min": timers[0],
                "avg": timers[1],
                "max": timers[2],
                "stddev": timers[3]
            }
        except:  # noqa: E722
            data = {
                "host": ip,
                "success": result.returncode == 0,
                "stdout": result.stdout,
            }
        return data

    @staticmethod
    def ping(hosts=None, count=1, processors=3):
        """
        ping a list of given ip addresses

        :param hosts: a list of ip addresses
        :param count: number of pings to run per ip
        :param processors: number of processors to Pool
        :return: list of dicts representing the ping result
        """

        # first expand the ips to a list
        hosts = Parameter.expand(hosts)

        # wrap ip and count into one list to be sent to Pool map
        args = [{'ip': ip, 'count': count} for ip in hosts]

        with Pool(processors) as p:
            res = p.map(Host._ping, args)
            p.close()
            p.join()

        return res

    @staticmethod
    def ssh_keygen(hosts=None,
                   filename="~/.ssh/id_rsa",
                   username=None,
                   processors=3,
                   dryrun=False,
                   verbose=True):
        """
        generates the keys on the specified hosts.
        this fonction does not work well as it still will aski if we overwrite.

        :param hosts:
        :param filename:
        :param username:
        :param output:
        :param dryrun:
        :param verbose:
        :return:
        """
        hosts = Parameter.expand(hosts)
        command = f'ssh-keygen -q -N "" -f {filename} <<< y'
        result_keys = Host.ssh(hosts=hosts,
                               command=command,
                               username=username,
                               dryrun=dryrun,
                               processors=processors,
                               executor=os.system)
        Host._print(result_keys)
        result_keys = Host.ssh(hosts=hosts,
                               processors=processors,
                               command='cat .ssh/id_rsa.pub',
                               username=username)
        return result_keys

    @staticmethod
    def gather_keys(username=None,
                    hosts=None,
                    filename="~/.ssh/id_rsa.pub",
                    key="~/.ssh/id_rsa",
                    processors=3,
                    dryrun=False):
        """
        returns in a list the keys of the specified hosts

        :param username:
        :param hosts:
        :param filename:
        :param key:
        :param dryrun:
        :return:
        """
        names = Parameter.expand(hosts)

        results_key = Host.ssh(hosts=names,
                               command='cat ~/.ssh/id_rsa.pub',
                               username=username,
                               verbose=False)
        # results_authorized = Host.ssh(hosts=names,
        #                              command='cat .ssh/id_rsa.pub',
        #                              username=username,
        #                              verbose=False)

        filename = path_expand(filename)
        content = readfile(filename).strip()
        localkey = {
            'host': "localhost",
            'command': [''],
            'execute': "",
            'stdout': content,
            'stderr': None,
            'returncode': True,
            'success': True,
            'date': DateTime.now()
        }
        if results_key is None:  # and results_authorized is None:
            return ""

        # getting the output and also removing duplicates

        keys = [localkey['stdout'].strip()]
        for element in results_key:
            if element['stdout'] != '':
                keys.append(element['stdout'].strip())

        output = list(set(keys))

        output = '\n'.join(output)
        print(output)

        return output



import random
import string


def generate_strong_pass():
    """
    Generates a password from letters, digits and punctuation

    :return: password
    :rtype: str
    """
    length = random.randint(10, 15)
    password_characters = \
        string.ascii_letters + \
        string.digits + \
        string.punctuation
    return ''.join(random.choice(password_characters) for i in range(length))



import os
import os.path
from builtins import bytes

import oyaml as yaml


#
# TODO: this does not follow our conventions, should this not be done at __init__?
#
###############################################################
#  make yaml understand unicode

def yaml_construct_unicode(self, node):
    return self.construct_scalar(node)


yaml.Loader.add_constructor(u'tag:yaml.org,2002:python/unicode', yaml_construct_unicode)
yaml.SafeLoader.add_constructor(u'tag:yaml.org,2002:python/unicode', yaml_construct_unicode)


###############################################################
#  the db api

class YamlDB(object):
    """A YAML-backed Key-Value database to store strings
    """

    def __init__(self, path):
        self._db = dict()

        self.path = path

        prefix = os.path.dirname(self.path)
        if not os.path.exists(prefix):
            os.makedirs(prefix)

        if os.path.exists(self.path):
            with open(self.path, 'rb') as dbfile:
                self._db = yaml.safe_load(dbfile) or dict()

        self.flush()

    def flush(self):
        string = yaml.dump(self._db, default_flow_style=False)
        bits = bytes(string, encoding='utf-8')
        with open(self.path, 'wb') as dbfile:
            dbfile.write(bits)

    def __setitem__(self, k, v):
        self._db[str(k)] = str(v)
        self.flush()

    def __getitem__(self, k):
        return self._db[k]

    def __delitem__(self, k):
        del self._db[k]
        self.flush()

    def __contains__(self, k):
        return k in self._db

    def __iter__(self):
        return iter(self._db)

    def __len__(self):
        return len(self._db)

    def close(self):
        """This is a NoOP for backwards compatibility"""
        pass

    def clear(self):
        """Truncate the database"""
        self._db.clear()
        self.flush()

    def set(self, **kwargs):
        for name in kwargs:
            self._db[name] = kwargs[name]

    def dict(self):
        return self._db

    def get(self, key, value=None):
        self._db.get(key, value)



#
# use full width of jupyter notebook windows
#

try:
    __IPYTHON__
    from IPython.core.display import display, HTML
    display(HTML("<style>.container { width:100% !important; }</style>"))

    from pprint import pprint  # noqa: F401

    from cloudmesh.common.util import banner

    banner("Cloudmesh")

except NameError:
    pass



from cloudmesh.common.variables import Variables
from cloudmesh.common.util import banner
from pprint import pformat
import inspect
import os
import threading
import traceback

verbose_lock = threading.Lock()


def TRACE():
    """
    prints the current trace stack
    """
    traceback.print_stack()


def tracefunc(frame, event, arg, indent=[0]):
    if event == "call":
        indent[0] += 2
        print("-" * indent[0] + "> call function", frame.f_code.co_name)
    elif event == "return":
        print("<" + "-" * indent[0], "exit function", frame.f_code.co_name)
        indent[0] -= 2
    return tracefunc


# sys.settrace(tracefunc)


# noinspection PyPep8Naming
def VERBOSE(msg, label=None, color="BLUE", verbose=9, location=True,
            secrets=["OS_PASSWORD",
                     "OS_USERNAME",
                     "client_secret",
                     "client_id",
                     "project_id",
                     "AZURE_TENANT_ID",
                     "AZURE_SUBSCRIPTION_ID",
                     "AZURE_APPLICATION_ID",
                     "AZURE_SECRET_KEY: TBD",
                     "EC2_ACCESS_ID: TBD",
                     "EC2_SECRET_KEY",
                     "MONGO_PASSWORD"]
            ):
    """
    Prints a data structure in verbose mode

    :param msg: the msg to be printed, can be a data structure such as a dict
    :param label: the  label to be used, defaults to the name of the msg variable
    :param color: the color
    :param verbose: indicates when to print it. If verbose in cloudmesh is
                    higher than the specified value it is printed
    :return:
    """

    _verbose = int(Variables()["verbose"] or 0)
    if _verbose >= verbose:

        verbose_lock.acquire()

        if label is None:
            label = inspect.stack()[1][4][0].strip().replace("VERBOSE(", "")
            label = label.split(",")[0][:-1]

        if location:
            cwd = os.getcwd()
            frame = inspect.stack()[1]
            module = inspect.getmodule(frame[0])
            filename = module.__file__.replace(cwd, ".")
            lineno = str(inspect.stack()[1][2]) + ":" + str(inspect.stack()[1][3])
            # print("# FILENAME:", filename, sep="")

        hline = "\n" + 70 * "-" + "\n"

        if type(msg) == dict and secrets is not None:
            tmp = dict(msg)
            for key in secrets:
                if key in tmp:
                    tmp[key] = "********"
            banner(lineno + " " + filename + hline + pformat(tmp),
                   label=label,
                   color=color)
        else:
            banner(lineno + " " + filename + hline + pformat(msg),
                   label=label,
                   color=color)

        verbose_lock.release()



"""
Implementation of a function the set the WIFI configuration.
This function is primarily developed for a Raspberry PI
"""
import textwrap

from cloudmesh.common.console import Console
from cloudmesh.common.util import writefile
from cloudmesh.common.sudo import Sudo


class Wifi:
    """
    The class is used to group a number of useful variables and functions so
    it is easier to program and manage Wifi configurations.

    The default location for the configuration file is

    /etc/wpa_supplicant/wpa_supplicant.conf

    """

    location = "/etc/wpa_supplicant/wpa_supplicant.conf"

    template_key = textwrap.dedent("""
        ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev
        update_config=1
        country={country}

        network={{
                ssid="{ssid}"
                key_mgmt=NONE
        }}
    """)

    template_psk = textwrap.dedent("""
        ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev
        update_config=1
        country=US

        network={{
                ssid="{ssid}"
                psk="{password}"
                key_mgmt=WPA-PSK
        }}
    """)  # noqa: W293

    template = template_psk

    @staticmethod
    def set(ssid=None,
            password=None,
            country="US",
            psk=True,
            location=location,
            sudo=False):
        """
        Sets the wifi. Only works for psk based wifi

        :param ssid: The ssid
        :type ssid: str
        :param password: The password
        :type password: str
        :param country: Two digit country code
        :type country: str
        :param psk: If true uses psk authentication
        :type psk: bool
        :param location: The file where the configuration file should be written to
        :type location: str
        :param sudo: If tru the write will be done with sudo
        :type sudo: bool
        :return: True if success
        :rtype: bool
        """

        if ssid is None or (psk and password is None):
            Console.error("SSID or password not set")
            return False

        if psk:
            config = Wifi.template.format(**locals())
        else:
            config = Wifi.template_key.format(**locals())
        try:
            if sudo:
                Sudo.writefile(location, config)
            else:
                writefile(location, config)

        except FileNotFoundError as e:  # noqa: F841
            Console.error(f"The file does not exist: {location}")
            return False
        return True



import platform
import sys
import os
import socket
from pathlib import Path
from cloudmesh.common.util import readfile
from cloudmesh.common.util import is_gitbash
from cloudmesh.common.util import is_powershell
from cloudmesh.common.util import is_cmd_exe
from cloudmesh.common.util import is_local
from collections import OrderedDict
import pip
import psutil
import humanize
import re
import multiprocessing
from cloudmesh.common.DateTime import DateTime


def os_is_windows():
    """
    Checks if the os is windows

    :return: True is windows
    :rtype: bool
    """
    return platform.system() == "Windows"


# noinspection PyBroadException
def os_is_linux():
    """
    Checks if the os is linux

    :return: True is linux
    :rtype: bool
    """
    try:
        content = readfile('/etc/os-release')
        return platform.system() == "Linux" and "raspbian" not in content
    except:  # noqa: E722
        return False


def os_is_mac():
    """
    Checks if the os is macOS

    :return: True is macOS
    :rtype: bool
    """
    return platform.system() == "Darwin"


# noinspection PyBroadException
def os_is_pi():
    """
    Checks if the os is Raspberry OS

    :return: True is Raspberry OS
    :rtype: bool
    """
    try:
        content = readfile('/etc/os-release')
        return platform.system() == "Linux" and "raspbian" in content
    except:  # noqa: E722
        return False


def has_window_manager():
    if os_is_mac() or os_is_windows():
        return True
    else:
        return "GNOME_TERMINAL_SCREEN" in os.environ or "GNOME_TERMINAL_SERVICE" in os.environ


def sys_user():
    """
    TODO

    Returns:

    """
    if "COLAB_GPU" in os.environ:
        return "collab"
    try:
        if sys.platform == "win32":
            return os.environ["USERNAME"]
    except:  # noqa: E722
        pass
    try:
        return os.environ["USER"]
    except:  # noqa: E722
        pass
    try:
        if os.environ["HOME"] == "/root":
            return "root"
    except:  # noqa: E722
        pass

    return "None"


def get_platform():
    """
    TODO

    Returns:

    """
    if sys.platform == "darwin":
        return "macos"
    elif sys.platform == "win32":
        return "windows"
    try:
        content = readfile('/etc/os-release')
        if sys.platform == 'linux' and "raspbian" in content:
            return "raspberry"
        else:
            return sys.platform
    except:  # noqa: E722
        return sys.platform


def systeminfo(info=None, user=None, node=None):
    """
    TODO

    Args:
        info (TODO):  TODO
        user (str): TODO
        node (str): TODO

    Returns: dict of the info

    """
    uname = platform.uname()
    mem = psutil.virtual_memory()

    # noinspection PyPep8
    def add_binary(value):
        try:
            r = humanize.naturalsize(value, binary=True)
        except:  # noqa: E722
            r = ""
        return r

    try:
        frequency = psutil.cpu_freq()
    except:  # noqa: E722
        frequency = None

    try:
        cores = psutil.cpu_count(logical=False)
    except:  # noqa: E722
        cores = "unkown"

    operating_system = get_platform()

    description = ""
    try:
        if operating_system == "macos":
            description = os.popen("sysctl -n machdep.cpu.brand_string").read()
        elif operating_system == "win32":
            description = platform.processor()
        elif operating_system == "linux":
            lines = readfile("/proc/cpuinfo").strip().splitlines()
            for line in lines:
                if "model name" in line:
                    description = re.sub(".*model name.*:", "", line, 1)
    except:  # noqa: E722
        pass

    data = OrderedDict({
        'cpu': description.strip(),
        'cpu_count': multiprocessing.cpu_count(),
        'cpu_threads': multiprocessing.cpu_count(),
        'cpu_cores': cores,
        'uname.system': uname.system,
        'uname.node': uname.node,
        'uname.release': uname.release,
        'uname.version': uname.version,
        'uname.machine': uname.machine,
        'uname.processor': uname.processor,
        'sys.platform': sys.platform,
        'python': sys.version,
        'python.version': sys.version.split(" ", 1)[0],
        'python.pip': pip.__version__,
        'user': sys_user(),
        'mem.percent': str(mem.percent) + " %",
        'frequency': frequency
    })
    for attribute in ["total",
                      "available",
                      "used",
                      "free",
                      "active",
                      "inactive",
                      "wired"
                      ]:
        try:
            data[f"mem.{attribute}"] = \
                humanize.naturalsize(getattr(mem, attribute), binary=True)
        except:  # noqa: E722
            pass
    # svmem(total=17179869184, available=6552825856, percent=61.9,

    if data['sys.platform'] == 'darwin':
        data['platform.version'] = platform.mac_ver()[0]
    elif data['sys.platform'] == 'win32':
        data['platform.version'] = platform.win32_ver()
    else:
        data['platform.version'] = uname.version

    try:
        release_files = Path("/etc").glob("*release")
        for filename in release_files:
            content = readfile(filename.resolve()).splitlines()
            for line in content:
                if "=" in line:
                    attribute, value = line.split("=", 1)
                    attribute = attribute.replace(" ", "")
                    data[attribute] = value
    except:  # noqa: E722
        pass
    if info is not None:
        data.update(info)
    if user is not None:
        data["user"] = user
    if node is not None:
        data["uname.node"] = node
    data["date"] = str(DateTime.now())
    return dict(data)



# see https://stackoverflow.com/questions/2536307/decorators-in-the-python-standard-lib-deprecated-specifically

import functools
import inspect
import warnings
from cloudmesh.common.console import Console

string_types = (type(b''), type(u''))


def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, string_types):

        # The @deprecated is used with a 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated("please, use another function")
        #    def old_function(x, y):
        #      pass

        def decorator(func1):

            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter('always', DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2
                )
                warnings.simplefilter('default', DeprecationWarning)
                Console.warning(fmt1.format(name=func1.__name__, reason=reason))
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):

        # The @deprecated is used without any 'reason'.
        #
        # .. code-block:: python
        #
        #    @deprecated
        #    def old_function(x, y):
        #      pass

        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2
            )
            warnings.simplefilter('default', DeprecationWarning)
            Console.warning(fmt2.format(name=func2.__name__, reason=reason))
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))



from cloudmesh.common.parameter import Parameter
from cloudmesh.common.util import path_expand
from cloudmesh.common.strdb import YamlDB
from cloudmesh.common.console import Console


class Variables(object):
    def __init__(self, filename=None):
        self.filename = path_expand(filename or "~/.cloudmesh/variables.dat")
        self.data = YamlDB(self.filename)

    def get(self, key, default=None):
        return self.data.get(key, default)

    def __getitem__(self, key):
        if key not in self.data:
            return None
        else:
            return self.data[key]

    def __setitem__(self, key, value):
        # print("set", key, value)
        self.data[str(key)] = value

    def __delitem__(self, key):
        if key in self.data:
            del self.data[str(key)]

    def __contains__(self, item):
        return str(item) in self.data

    def __str__(self):
        return str(self.data)

    def __len__(self):
        return len(self.data)

    def __add__(self, directory):
        for key in directory:
            self.data[key] = directory[key]

    def __sub__(self, keys):
        for key in keys:
            del self.data[key]

    def __iter__(self):
        return iter(self.data)

    def close(self):
        self.data.close()

    def clear(self):
        self.data.clear()

    def dict(self):
        return self.data._db

    def parameter(self, attribute, position=0):
        value = str(self.data[attribute])
        expand = Parameter.expand(value)[position]
        return expand

    def boolean(self, key, value):
        if str(value).lower() in ["true", "on"]:
            self.data[str(key)] = True
        elif str(value).lower() in ["false", "off"]:
            self.data[str(key)] = False
        else:
            Console.error("Value is not boolean")


if __name__ == "__main__":
    v = Variables()
    print(v)

    v["gregor"] = "gregor"
    assert "gregor" in v
    del v["gregor"]
    assert "gregor" not in v



"""
Class for starting and stopping named timers.

Here is a simple example. the user and node parameters can be omitted, but they help in case you like
to change the system retrieved values.

from pprint import pprint
from cloudmesh.common.StopWatch import StopWatch

import time

user="gregor"
node="5950x"

for t in ["a", "b"]:
    StopWatch.start(t)
    time.sleep(0.1)
    StopWatch.stop(t)

pprint(StopWatch.get_benchmark(user=user, node=node))

StopWatch.benchmark(user=user, node=node)

## Context Block

StoWatch alos comes with a context block allowing to use the convenient with statement. However thi sis only recommended
when the block is small as it is easy to lose the indentation in larger code. You can also certainly split the code up in
functions so it is easier managable.

Here is a simple example on how to use the with statement, showing that is can be used for streams and files.
The mode tetermins if the file will be recreated, or if it will be appanded on. In addition there is the ability to
write metadata into the record with a dict that can optionally be passed along.

In order not to overwrite the value of an event, you must give it a unique name.

    from cloudmesh.common.StopWatch import StopWatchBlock
    from cloudmesh.common.StopWatch import StopWatch
    from cloudmesh.common.util import readfile

    import time

    data = {"step": "value"}

    StopWatch.event("event-start")
    d = {"key": "value"}
    StopWatch.event("event-with-value", d)

    with StopWatchBlock("total"):
        time.sleep(1.0)

    with StopWatchBlock("dict", data=data):
        time.sleep(1.0)
        data["step"] = 1

    with StopWatchBlock("file", data=data, log="a.log", mode="w"):
        time.sleep(1.0)
        data["step"] = 2

    with StopWatchBlock("append", data=data, log="a.log", mode="a"):
        time.sleep(1.0)
        data["step"] = 3

    content = readfile ("a.log")
    print (79*"=")
    print (content.strip())
    print (79*"=")


    StopWatch.event("event-stop")

    StopWatch.benchmark(sysinfo=False,
                        tag="myexperiment",
                        user="gregor",
                        node="computer",
                        attributes=["timer", "status", "time", "start", "tag", "msg"]

    # for more examples, see our pytest:
    # * <https://github.com/cloudmesh/cloudmesh-common/blob/main/tests/test_stopwatch.py>
    #
    # there we demonstrate how to timers in a loop as individual timers and as sum.
    # We also showcase how to add a message to timers
)

Integration with MLPerf Logging

To also produce output that conforms to MLPerf, cloudmesh. STopWatch will detect if you have mperf logging installed.
We recommend tho install the newest version as follows


::
    git clone https://github.com/mlperf/logging.git mlperf-logging
    cd mlperf-logging
    pip install -e .

Now you can just use the STopwatch as before.

We will add here aditional information, such as setting up the configuration for mlperf logging

"""
import os
import platform
import time
import datetime
import pprint
import pathlib
import yaml
import sys

from cloudmesh.common.console import Console
from cloudmesh.common.Tabulate import Printer
from cloudmesh.common.systeminfo import systeminfo as cm_systeminfo
from cloudmesh.common.util import writefile
from cloudmesh.common.util import readfile
from cloudmesh.common.DateTime import DateTime

from time import perf_counter


def rename(newname):
    """
    decorator to rename a function
    :param newname: function name
    :type newname: str
    :return: renamed function
    :rtype: object
    """

    def decorator(f):
        f.__name__ = newname
        return f

    return decorator


def benchmark(func):
    """
    decorator to benchmark a function
    :param func: function
    :type func: object
    :return: function with benchmarks based on the name of the function
    :rtype: object
    """

    @rename(func.__name__)
    def wrapper(*args, **kwargs):
        StopWatch.start(func.__name__)
        func(*args, **kwargs)
        StopWatch.stop(func.__name__)

    return wrapper


def import_mllog():
    try:
        from mlperf_logging import mllog
    except:  # noqa: E722
        Console.error("You need to install mllogging to use it")
        sys.exit()
    return mllog


class StopWatch(object):
    """
    A class to measure times between events.
    """
    debug = False
    verbose = True
    # Timer start dict
    timer_start = {}
    # Timer end dict
    timer_end = {}
    # Timer diff
    timer_elapsed = {}
    # records a status
    timer_status = {}
    # records a dt
    timer_sum = {}
    # msg
    timer_msg = {}
    # mllogger
    timer_values = {}
    mllogging = False
    mllogger = None

    @classmethod
    def activate_mllog(cls, filename="cloudmesh_mllog.log", config=None, stack_offset=1):
        # global mllog
        cls._mllog_import = import_mllog()

        if config is None:
            cms_mllog = dict(
                default_namespace="cloudmesh",
                default_stack_offset=stack_offset,
                default_clear_line=False
            )
        else:
            cms_mllog = config

        cls.mllogging = True
        cls.mllogger = cls._mllog_import.get_mllogger()
        cls._mllog_import.config(filename=filename)
        cls._mllog_import.config(**cms_mllog
            # useful when refering to linenumbers in separate code
            # root_dir=os.path.normpath(
            #    os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", ".."))
        )

    # @classmethod
    # def progress(cls, percent, status="running", pid=None):
    #     if pid is None:
    #         pid = os.getpid()
    #     if "SLURM_JOB_ID" in os.environ:
    #         pid = os.environ["SLURM_JOB_ID"] #TODO - may need to be updated (monitor of long running jobs)
    #     print(f"# cloudmesh status={status} progress={percent} pid={pid}")

    @classmethod
    def progress(cls, percent, status="running", pid=None, variable=None):
        """Prints progress of an event, recording against a pid and providing additional variable.

        :percent: 0-100 value
        :status: Message to associate to the recording, default - running
        :pid: The associated Process ID for this event.
        :variable: Any valid python type with a __str__ method.

        :returns: The progress message as a string
        """
        if pid is None:
            pid = os.getpid()
        if "SLURM_JOB_ID" in os.environ:
            # TODO - may need to be updated (monitor of long running jobs)
            pid = os.environ["SLURM_JOB_ID"]
        msg = f"# cloudmesh status={status} progress={percent} pid={pid}"
        if variable is not None:
            msg = msg + f" variable={variable}"
        print(msg)
        return msg
        try:
            config = yaml.safe_load(readfile(configfile).strip())
        except:  # noqa: E722
            config = {
                "benchmark": {}
            }
        config["benchmark"].update(argv)


    @classmethod
    def organization_mllog(cls, configfile=None, prefix_ = 'benchmark', flatdict_ = False, **argv):
        try:
            from mlperf_logging import mllog
        except Exception:  # noqa: E722
            Console.error("You need to install mllogging to use it")
            sys.exit()

        try:
            with open(pathlib.Path(configfile), 'r') as stream:
                _config = yaml.safe_load(stream)
        except Exception as e:  # noqa: E722
            _config = {
                "benchmark": {}
            }
        if flatdict_:
            prefix=f"{prefix_}"
            for k,v in argv.items():
                _config[f"{prefix}.{k}"] = v
        else:
            _config.update(argv)

        print (_config)

        for key, attribute in [
            (mllog.constants.SUBMISSION_BENCHMARK, 'name'),
            (mllog.constants.SUBMISSION_POC_NAME, 'user'),
            (mllog.constants.SUBMISSION_POC_EMAIL, 'e-mail'),
            (mllog.constants.SUBMISSION_ORG, 'organisation'),
            (mllog.constants.SUBMISSION_DIVISION, 'division'),
            (mllog.constants.SUBMISSION_STATUS, 'status'),
            (mllog.constants.SUBMISSION_PLATFORM, 'platform')
            ]:
            try:
                if flatdict_:
                    cls.mllogger.event(key, value=_config[f"{prefix}.{attribute}"])
                else:
                    cls.mllogger.event(key, value=_config["benchmark"][attribute])
            except AttributeError as e:
                print(f"Missing/invalid standard property {key}")


    @classmethod
    def keys(cls):
        """returns the names of the timers"""
        return list(cls.timer_end.keys())

    @classmethod
    def status(cls, name, value):
        """
        starts a timer with the given name.

        :param name: the name of the timer
        :type name: string
        :param value: value of the nameed of a status
        :type value: bool

        """
        if cls.debug:
            print("Timer", name, "status", value)
        cls.timer_status[name] = value

    @classmethod
    def get_message(cls, name):
        """
        starts a timer with the given name.

        :param name: the name of the timer
        :type name: string

        """
        return cls.timer_msg[name]

    @classmethod
    def message(cls, name, value):
        """
        starts a timer with the given name.

        :param name: the name of the timer
        :type name: string
        :param value: the value of the message
        :type value: bool

        """
        cls.timer_msg[name] = value

    @classmethod
    def event(cls, name, msg=None, values=None, suppress_stopwatch=False, suppress_mllog=False):
        """
        Adds an event with a given name, where start and stop is the same time.

        :param name: the name of the timer
        :type name: string
        :param msg: a message to attach to this event
        :type msg: string
        :param values: data that is associated with the event that is converted
                       to a string
        :type values: object
        :param suppress_stopwatch: suppresses executing any stopwatch code.
                                Useful when only logging an mllog event.
        :type suppress_stopwatch: bool
        :param suppress_mllog: suppresses executing any mllog code.  Useful
                               when only interacting with stopwatch timers.
        :type suppress_mllog: bool

        :returns: None
        :rtype: None
        """
        if not suppress_stopwatch:
            StopWatch.start(name, suppress_mllog=True)
            StopWatch.stop(name, suppress_mllog=True)
            StopWatch.timer_end[name] = StopWatch.timer_start[name]
            if values:
                StopWatch.timer_values[name] = values

            if msg is not None:
                StopWatch.message(name, str(msg))
        if cls.mllogging and not suppress_mllog:
            if values is not None:
                cls.mllogger.event(key=name, value=str(values), stack_offset=2)
            else:
                cls.mllogger.event(key=name, stack_offset=2)

    @classmethod
    def log_event(cls, **kwargs):
        """Logs an event using the passed keywords as parameters to be logged,
           prefiltered by mlperf_logging's standard api.

        :param kwargs: an unpacked dictionary of key=value entries to be
                       leveraged when logging an event to both the cloudmesh
                       stopwatch and mllog.  If the keyword matches
                       mlperf_logging's constants, the value will be replaced
                       with the standardized string
        :type kwargs: dict
        """
        for key, value in kwargs.items():
            mlkey = cls.mllog_lookup(key)
            cls.event(mlkey, msg=mlkey, values=value)


    @classmethod
    def log_constant(cls, **kwargs):
        """Deprecated.  Use `log_event`."""
        cls.log_event(**kwargs)

    @classmethod
    def mllog_lookup(cls, key: str) -> str:
        """Performs a dynamic lookup for the string representation of a
           mlperf constant.  If the value isn't found, it will return a string
           of the pattern mllog-event-{key}

        :param key: The name of the constant to look up
        :type key: string

        :returns: The decoded value of the constant.
        :rtype: string
        """
        try:
            from mlperf_logging.mllog import constants as mlconst
        except ImportError as e:
            Console.error("You need to install mlperf_logging to use it")
            raise e
        try:
            key_str = getattr(mlconst, key)
        except AttributeError as e:
            key_str = f"mllog-event-{key}"
        return key_str


    @classmethod
    def start(cls, name, values=None, mllog_key=None, suppress_stopwatch=False, suppress_mllog=False):
        """
        starts a timer with the given name.

        :param name: the name of the timer
        :type name: string
        :param values: any python object with a __str__ method to record with
                       the event.
        :type values: object
        :param mllog_key: Specifies the string name of an mllog constant to
                          associate to this timer start.  If no value is passed
                          and mllogging is enabled, then `name` is used.
        :type mllog_key: string
        :param suppress_stopwatch: When true, prevents all traditional
                                   stopwatch logic from running.  This is
                                   useful when attempting to interact with
                                   mllog-only.
        :type suppress_stopwatch: bool
        :param suppress_mllog: When true, prevents all mllog events from
                               executing.  Useful when working with stopwatch
                               timers-only.

        :returns: None
        :rtype: None
        """
        if not suppress_stopwatch:
            if cls.debug:
                print("Timer", name, "started ...")
            if name not in cls.timer_sum:
                cls.timer_sum[name] = 0.0
            cls.timer_start[name] = time.time()
            cls.timer_end[name] = None
            cls.timer_status[name] = None
            cls.timer_msg[name] = None
            if values:
                StopWatch.timer_values[name] = values

        if cls.mllogging and not suppress_mllog:
            if mllog_key is None:
                key = name
            else:
                key = cls.mllog_lookup(mllog_key)
            if values is not None:
                cls.mllogger.start(key=key, value=str(values), stack_offset=2)
            else:
                cls.mllogger.start(key=key, stack_offset=2)

    @classmethod
    def stop(cls, name, state=True, values=None, mllog_key=None, suppress_stopwatch=False, suppress_mllog=False):
        """
        stops the timer with a given name.

        :param name: the name of the timer
        :type name: string
        :param state: When true, updates the status of the timer.
        :type state: bool
        :param mllog_key: Specifies the string name of an mllog constant to
                          associate to this timer start.  If no value is passed
                          and mllogging is enabled, then `name` is used.
        :type mllog_key: string
        :param suppress_stopwatch: When true, prevents all traditional
                                   stopwatch logic from running.  This is
                                   useful when attempting to interact with
                                   mllog-only.
        :type suppress_stopwatch: bool
        :param suppress_mllog: When true, prevents all mllog events from
                               executing.  Useful when working with stopwatch
                               timers-only.

        :returns: None
        :rtype: None
        """
        if not suppress_stopwatch:
            cls.timer_end[name] = time.time()
            # if cumulate:
            #    cls.timer_end[name] = cls.timer_end[name] + cls.timer_last[name]
            cls.timer_sum[name] = cls.timer_sum[name] + cls.timer_end[name] - cls.timer_start[name]
            cls.timer_status[name] = state
            if values:
                StopWatch.timer_values[name] = values

        if cls.mllogging and not suppress_mllog:
            if mllog_key is None:
                key = name
            else:
                key = cls.mllog_lookup(mllog_key)
            if values is not None:
                cls.mllogger.end(key=key, value=str(values), stack_offset=2)
            else:
                cls.mllogger.end(key=key, stack_offset=2)

        if cls.debug and not suppress_stopwatch:
            print("Timer", name, "stopped ...")

    @classmethod
    def get_status(cls, name):
        """
        sets the status of the timer with a given name.

        :param name: the name of the timer
        :type name: string
        """
        return cls.timer_status[name]

    # noinspection PyPep8
    @classmethod
    def get(cls, name, digits=4):
        """
        returns the time of the timer.

        :param name: the name of the timer
        :type name: string
        :rtype: the elapsed time
        """
        if name in cls.timer_end:
            try:
                diff = cls.timer_end[name] - cls.timer_start[name]
                if round is not None:
                    cls.timer_elapsed[name] = round(diff, digits)
                else:
                    cls.timer_elapsed[name] = diff
                return cls.timer_elapsed[name]
            except:  # noqa: E722
                return None
        else:
            return "undefined"

    @classmethod
    def sum(cls, name, digits=4):
        """
        returns the sum of the timer if used multiple times

        :param name: the name of the timer
        :type name: string
        :rtype: the elapsed time
        """
        if name in cls.timer_end:
            try:
                diff = cls.timer_sum[name]
                if round is not None:
                    return round(diff, digits)
                else:
                    return diff
            except:  # noqa: E722
                return None
        else:
            return "undefined"

    @classmethod
    def clear(cls):
        """
        clear start and end timer_start
        """
        cls.timer_start.clear()
        cls.timer_end.clear()
        cls.timer_sum.clear()
        cls.timer_status.clear()
        cls.timer_elapsed.clear()
        cls.timer_msg.clear()

    @classmethod
    def print(cls, *args):
        """
        prints a timer. The first argument is the label if it exists, the last is the timer
        :param args: label, name
        :return:
        """
        if cls.verbose:
            if len(args) == 2:
                print(args[0], str("{0:.2f}".format(cls.get(args[1]))), "s")
            else:
                raise Exception("StopWatch: wrong number of arguments")

    @classmethod
    def output(cls, name):
        """
        prints a timer. The first argument is the label if it exists, the last is the timer
        :param args: label, name
        :return:
        """
        print(name, str("{0:.2f}".format(cls.get(name))), "s")

    @classmethod
    def __str__(cls):
        """
        returns the string representation of the StopWatch
        :return: string of the StopWatch
        :rtype: str
        """
        s = ""
        for t in cls.timer_end:
            data = {"label": t,
                    "start": str(cls.timer_start[t]),
                    "end": str(cls.timer_end[t]),
                    "status": str(cls.timer_status[t]),
                    "elapsed": str(cls.get(t)),
                    "newline": os.linesep}
            s += "{label} {start} {end} {elapsed} {status} {newline}".format(
                **data)
        return s

    @classmethod
    def systeminfo(cls, data=None):
        """
        Print information about the system

        :param data: additional data to be integrated
        :type data: dict
        :return: a table with data
        :rtype: str
        """
        data_platform = cm_systeminfo()
        if data is not None:
            data_platform.update(data)
        return Printer.attribute(
            data_platform,
            order=["Machine Attribute", "Value"],
            output="table"
        )

    @classmethod
    def get_sysinfo(cls,
                    node=None,
                    user=None):
        data_platform = cm_systeminfo(node=node, user=user)
        return data_platform

    @classmethod
    def get_benchmark(cls,
                      sysinfo=True,
                      tag=None,
                      node=None,
                      user=None,
                      total=False,
                      ):
        """
        prints out all timers in a convenient benchmark table

        :param sysinfo: controls if system info shoul be printed.
        :type sysinfo: bool
        :param csv: contols if the data should be printed also as csv strings
        :type csv: bool
        :param prefix: The prefix used for the csv string
        :type prefix: str
        :param tag: overwrites the tag
        :type tag: str
        :param sum: prints the sums (not used)
        :type sum: bool
        :param node: overwrites the name of the node
        :type node: str
        :param user: overwrites the name of the user
        :type user: str
        :param attributes: list of additional attributes to print
        :type attributes: list
        :return: prints the information
        :rtype: stdout
        """

        #
        # PRINT PLATFORM
        #

        data_platform = cm_systeminfo(user=user, node=node)
        if sysinfo:
            print(Printer.attribute(
                data_platform,
                output="table"
            ))

        benchmark_data = {
            'sysinfo': data_platform,
        }

        #
        # GET TIMERS
        #
        timers = StopWatch.keys()
        total_time = 0.0
        if len(timers) > 0:

            data_timers = {}
            for timer in timers:
                data_timers[timer] = {
                    'start': time.strftime("%Y-%m-%d %H:%M:%S",
                                           time.gmtime(
                                               StopWatch.timer_start[timer])),
                    'stop': time.strftime("%Y-%m-%d %H:%M:%S",
                                          time.gmtime(
                                              StopWatch.timer_end[timer])),
                    'time': StopWatch.get(timer, digits=3),
                    'sum': StopWatch.sum(timer, digits=3),
                    'status': StopWatch.get_status(timer),
                    'msg': StopWatch.get_message(timer),
                    'timer': timer,
                    'tag': tag or ''
                }
                total_time = total_time + StopWatch.get(timer)

            # print(Printer.attribute(data_timers, header=["Command", "Time/s"]))

            if 'benchmark_start_stop' in data_timers:
                del data_timers['benchmark_start_stop']

            for key in data_timers:
                if key != 'benchmark_start_stop' and data_timers[key]['status'] is None:
                    data_timers[key]['status'] = "failed"
                elif data_timers[key]['status'] is not None and data_timers[key]['status']:
                    data_timers[key]['status'] = "ok"

            if total:
                print("Total:", total_time)

            benchmark_data["benchmark"] = data_timers

        else:
            print("ERROR: No timers found")

        return benchmark_data

    @classmethod
    def benchmark(cls,
                  sysinfo=True,
                  timers=True,
                  csv=True,
                  prefix="# csv",
                  tag=None,
                  sum=True,
                  node=None,
                  user=None,
                  version=None,
                  attributes=None,
                  total=False,
                  filename=None):
        """
        prints out all timers in a convenient benchmark table

        :param sysinfo: controls if system info shoul be printed.
        :type sysinfo: bool
        :param csv: contols if the data should be printed also as csv strings
        :type csv: bool
        :param prefix: The prefix used for the csv string
        :type prefix: str
        :param tag: overwrites the tag
        :type tag: str
        :param sum: prints the sums (not used)
        :type sum: bool
        :param node: overwrites the name of the node
        :type node: str
        :param user: overwrites the name of the user
        :type user: str
        :param attributes: list of additional attributes to print
        :type attributes: list
        :return: prints the information
        :rtype: stdout
        """

        #
        # PRINT PLATFORM
        #
        content = "\n"

        data_platform = cm_systeminfo(user=user, node=node)
        if sysinfo:
            content = content + Printer.attribute(
                data_platform,
                order=["Machine Attribute", "Value"],
                output="table"
            )
            content = content + "\n"

        if timers:

            #
            # PRINT TIMERS
            #
            timers = StopWatch.keys()
            total_time = 0.0
            if len(timers) > 0:

                data_timers = {}
                for timer in timers:
                    data_timers[timer] = {
                        'start': time.strftime("%Y-%m-%d %H:%M:%S",
                                               time.gmtime(
                                                   StopWatch.timer_start[timer])),
                        'time': StopWatch.get(timer, digits=3),
                        'sum': StopWatch.sum(timer, digits=3),
                        'status': StopWatch.get_status(timer),
                        'msg': StopWatch.get_message(timer),
                        'timer': timer,
                        'tag': tag or ''
                    }
                    try:
                        total_time = total_time + StopWatch.get(timer)
                    except:  # noqa: E722
                        pass
                    for attribute in ["uname.node",
                                      "user",
                                      "uname.system",
                                      "uname.machine",
                                      "platform.version",
                                      "sys.platform"]:
                        if attribute == "user" and user is not None:
                            data_timers[timer][attribute] = user
                        elif attribute == "uname.node" and node is not None:
                            data_timers[timer][attribute] = node
                        else:
                            data_timers[timer][attribute] = data_platform[attribute]

                    if version is not None:
                        data_timers[timer]["platform.version"] = version


                # print(Printer.attribute(data_timers, header=["Command", "Time/s"]))

                if 'benchmark_start_stop' in data_timers:
                    del data_timers['benchmark_start_stop']

                for key in data_timers:
                    if key != 'benchmark_start_stop' and data_timers[key]['status'] is None:
                        data_timers[key]['status'] = "failed"
                    elif data_timers[key]['status'] is not None and data_timers[key]['status']:
                        data_timers[key]['status'] = "ok"

                if attributes is None:
                    order = [
                        "timer",
                        "status",
                        "time",
                        "sum",
                        "start",
                        "tag",
                        "msg",
                        "uname.node",
                        "user",
                        "uname.system",
                        "platform.version"
                    ]

                    header = [
                        "Name",
                        "Status",
                        "Time",
                        "Sum",
                        "Start",
                        "tag",
                        "msg",
                        "Node",
                        "User",
                        "OS",
                        "Version"
                    ]
                elif attributes == "short":
                    order = [
                        "timer",
                        "status",
                        "time"
                    ]

                    header = [
                        "Name",
                        "Status",
                        "Time"
                    ]
                else:
                    order = attributes
                    header = attributes
                content = content + "\n"
                content = content + Printer.write(
                    data_timers,
                    order=order,
                    header=header,
                    output="table"

                )

                if total:
                    content = content + f"Total: {total_time}"

                content = content + "\n"

                if csv:
                    if prefix is not None:
                        for entry in data_timers:
                            data_timers[entry]["# csv"] = prefix

                        order = ["# csv"] + order

                        content = content + Printer.write(
                            data_timers,
                            order=order,
                            header=header,
                            output="csv"
                        )
                    else:

                        content = content + pprint.pformat(data_timers, indent=4)
                        content = content + "\n"

                        content = content + Printer.write(
                            data_timers,
                            order=order[1:],
                            output="csv"
                        )
                        content = content + "\n"

            else:
                content = content + "ERROR: No timers found\n"

        print(content)
        if filename:
            writefile(filename, content)

    def load(filename,
             label=["name"], label_split_char=" ",
             attributes=['timer',
                         'status',
                         'time',
                         'sum',
                         'start',
                         'tag',
                         'msg',
                         'uname.node',
                         'user',
                         'uname.system',
                         'platform.version']):
        """
        Loads data written to a file from the #csv lines.
        If the timer name has spaces in it, it must also have a label tag in which each lable is the name when
        splitting up the timer name. The list of attributes is the list specified plus the once generated from the
        timer name by splitting.

        Example:
            data = StopWatch.load(logfile, label=["name", "n"], attributes=["timer", "time", "user", "uname.node"])


        :param label:
        :type label:
        :param attributes:
        :type attributes:
        :return:
        :rtype:
        """
        data = []
        headers = []
        content = readfile(logfile)
        lines = Shell.find_lines_with(content, what="# csv")
        data_attributes = lines[0].split(",")
        index_attributes = []
        for attribute in attributes:
            index_attributes.append(data_attributes.index(attribute))
        print(index_attributes)
        headers = attributes + label
        del lines[0]
        for line in lines:
            entry = line.split(",")
            entry = [entry[i] for i in index_attributes]
            label_tags = entry[0].split(label_split_char)
            entry = entry + label_tags
            data.append(entry)

        return {"headers": headers,
                "data": data}


class StopWatchBlock:

    def __init__(self, name, data=None, log=sys.stdout, mode="w"):
        self.name = name
        self.data = data
        self.log = log
        self.is_file = False
        self.start = datetime.datetime.now()
        if type(log) == str:
            self.is_file = True
            self.log = open(log, mode)

    def __enter__(self):
        StopWatch.start(self.name)
        return StopWatch.get(self.name)

    def __exit__(self, type, value, traceback):
        self.stop = datetime.datetime.now()
        StopWatch.stop(self.name)
        entry = StopWatch.get(self.name)
        if self.data:
            print(f"# {self.name}, {entry}, {self.start}, {self.stop}, {self.data}", file=self.log)
        else:
            print(f"# {self.name}, {entry}, {self.start}, {self.stop}", file=self.log)
        if self.is_file:
            self.log.close()



import subprocess
import collections
import glob
import inspect
import os
import random
import re
import shutil
import tempfile
import time
from contextlib import contextmanager
from getpass import getpass
import sys
import psutil
import requests
from pathlib import Path
from cloudmesh.common.console import Console
import pyfiglet
import socket
import platform

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


@contextmanager
def tempdir(*args, **kwargs):
    """A contextmanager to work in an auto-removed temporary directory

    Arguments are passed through to tempfile.mkdtemp

    example:

    >>> with tempdir() as path:
    ...    pass
    """

    d = tempfile.mkdtemp(*args, **kwargs)
    try:
        yield d
    finally:
        shutil.rmtree(d)


def check_root(dryrun=False, terminate=True):
    """
    check if I am the root user. If not, simply exits the program.
    :param dryrun: if set to true, does not terminate if not root user
    :type dryrun: bool
    :param terminate: terminates if not root user and dryrun is False
    :type terminate: bool
    """
    uid = os.getuid()
    if uid == 0:
        Console.ok("You are executing as a root user")
    else:
        Console.error("You do not run as root")
        if terminate and not dryrun:
            sys.exit()


def exponential_backoff(fn, sleeptime_s_max=30 * 60):
    """
    Calls `fn` until it returns True, with an exponentially increasing wait
    time between calls

    :param fn: the function to be called that returns Truw or False
    :type fn: object
    :param sleeptime_s_max: the sleep time in milliseconds
    :type sleeptime_s_max: int
    :return: None
    """
    sleeptime_ms = 500
    while True:
        if fn():
            return True
        else:
            print('Sleeping {} ms'.format(sleeptime_ms))
            time.sleep(sleeptime_ms / 1000.0)
            sleeptime_ms *= 2

        if sleeptime_ms / 1000.0 > sleeptime_s_max:
            return False


def download(source, destination, force=False):
    """
    Downloads the file from source to destination

    For large files, see cloudmesh.common.Shell.download

    :param source: The http source
    :param destination: The destination in the file system
    :param force: If True the file will be downloaded even if
                  it already exists
    """
    if os.path.isfile(destination) and not force:
        Console.warning(f"File {destination} already exists. "
                        "Skipping download ...")
    else:

        directory = os.path.dirname(destination)
        Path(directory).mkdir(parents=True, exist_ok=True)
        r = requests.get(source, allow_redirects=True)
        open(destination, 'wb').write(r.content)


def search(lines, pattern):
    """
    return all lines that match the pattern
    #TODO: we need an example

    :param lines:
    :param pattern:
    :return:
    """
    p = pattern.replace("*", ".*")
    test = re.compile(p)
    result = []
    for l in lines:  # noqa: E741
        if test.search(l):
            result.append(l)
    return result


def grep(pattern, filename):
    """Very simple grep that returns the first matching line in a file.
    String matching only, does not do REs as currently implemented.
    """
    try:
        # for line in file
        # if line matches pattern:
        #    return line
        return next((L for L in open(filename) if L.find(pattern) >= 0))
    except StopIteration:
        return ''


def is_local(host):
    """
    Checks if the host is the localhost

    :param host: The hostname or ip
    :return: True if the host is the localhost
    """
    return host in ["127.0.0.1",
                    "localhost",
                    socket.gethostname(),
                    # just in case socket.gethostname() does not work  we also try the following:
                    platform.node(),
                    socket.gethostbyaddr(socket.gethostname())[0]
                    ]


# noinspection PyPep8
def is_gitbash():
    """
    returns True if you run in a Windows gitbash

    :return: True if gitbash
    """
    try:
        exepath = os.environ['EXEPATH']
        return "Git" in exepath
    except:  # noqa: E722
        return False


def is_powershell():
    """
    True if you run in powershell

    :return: True if you run in powershell
    """
    # psutil.Process(parent_pid).name() returns -
    # cmd.exe for CMD
    # powershell.exe for powershell
    # bash.exe for git bash
    return (psutil.Process(os.getppid()).name() == "powershell.exe")


def is_cmd_exe():
    """
    return True if you run in a Windows CMD

    :return: True if you run in CMD
    """
    if is_gitbash():
        return False
    else:
        try:
            return os.environ['OS'] == 'Windows_NT'
        except:  # noqa: E722
            return False


def path_expand(text, slashreplace=True):
    """ returns a string with expanded variable.

    :param text: the path to be expanded, which can include ~ and environment variables
    :param text: string

    """
    result = os.path.expandvars(os.path.expanduser(text))

    if result.startswith("./"):
        result = result.replace(".", os.getcwd(), 1)

    if is_gitbash() or is_cmd_exe():
        if slashreplace:
            result = result.replace("/", "\\")

    return result


def convert_from_unicode(data):
    """
    converts unicode data to a string
    :param data: the data to convert
    :return:
    """
    # if isinstance(data, basestring):

    if isinstance(data, str):
        return str(data)
    elif isinstance(data, collectionsAbc.Mapping):
        return dict(map(convert_from_unicode, data.items()))
    elif isinstance(data, collectionsAbc.Iterable):
        return type(data)(map(convert_from_unicode, data))
    else:
        return data


def yn_choice(message, default='y', tries=None):
    """asks for a yes/no question.

    :param tries: the number of tries
    :param message: the message containing the question
    :param default: the default answer
    """
    # http://stackoverflow.com/questions/3041986/python-command-line-yes-no-input"""
    choices = 'Y/n' if default.lower() in ('y', 'yes') else 'y/N'
    if tries is None:
        choice = input(f"{message} ({choices}) ")
        values = ('y', 'yes', '') if default == 'y' else ('y', 'yes')
        return True if choice.strip().lower() in values else False
    else:
        while tries > 0:
            choice = input(f"{message} ({choices}) ('q' to discard)")
            choice = choice.strip().lower()
            if choice in ['y', 'yes']:
                return True
            elif choice in ['n', 'no', 'q']:
                return False
            else:
                print("Invalid input...")
                tries -= 1


def str_banner(txt=None, c="-", prefix="#", debug=True, label=None,
               color="BLUE", padding=False,
               figlet=False, font="big"):
    """
    prints a banner of the form with a frame of # around the txt::

    # --------------------------
    # txt
    # --------------------------

    :param color: prints in the given color
    :param label: adds a label
    :param debug: prints only if debug is true
    :param txt: a text message to be printed
    :type txt: string
    :param c: the character used instead of c
    :type c: character
    """
    output = ""
    if debug:
        output = "\n"
        output += prefix + " " + 70 * c + "\n"
        if padding:
            output += prefix + "\n"
        if label is not None:
            output += prefix + " " + label + "\n"
            output += prefix + " " + 70 * c + "\n"

        if txt is not None:

            if figlet:
                txt = pyfiglet.figlet_format(txt, font=font)

            for line in txt.splitlines():
                output += prefix + " " + line + "\n"
            if padding:
                output += prefix + "\n"
            output += prefix + " " + 70 * c + "\n"

    return output


def banner(txt=None, c="-", prefix="#", debug=True, label=None,
           color="BLUE", padding=False,
           figlet=False, font="big"):
    """
    prints a banner of the form with a frame of # around the txt::

    # --------------------------
    # txt
    # --------------------------

    :param color: prints in the given color
    :param label: adds a label
    :param debug: prints only if debug is true
    :param txt: a text message to be printed
    :type txt: string
    :param c: the character used instead of c
    :type c: character
    :param padding: ads additional comment line around the text so the banner is larger
    :type padding: bool
    """

    output = str_banner(txt=txt, c=c, prefix=prefix, debug=debug, label=label,
                        color=color, padding=padding, figlet=figlet, font=font)
    Console.cprint(color, "", output)


# noinspection PyPep8Naming
def HEADING(txt=None, c="#", color="HEADER"):
    """
    Prints a message to stdout with #### surrounding it. This is useful for
    pytests to better distinguish them.

    :param c: uses the given char to wrap the header
    :param txt: a text message to be printed
    :type txt: string
    """
    frame = inspect.getouterframes(inspect.currentframe())

    filename = frame[1][1].replace(os.getcwd(), "")
    line = frame[1][2] - 1
    method = frame[1][3]
    if txt is None:
        msg = "{} {} {}".format(method, filename, line)
    else:
        msg = "{}\n {} {} {}".format(txt, method, filename, line)

    print()
    banner(msg, c=c, color=color)


# noinspection PyPep8Naming
def FUNCTIONNAME():
    """
    Returns the anme of a function.
    """
    frame = inspect.getouterframes(inspect.currentframe())

    filename = frame[1][1].replace(os.getcwd(), "")
    line = frame[1][2] - 1
    method = frame[1][3]
    return method


def backup_name(filename):
    """
    :param filename: given a filename creates a backup name of the form
                     filename.bak.1. If the filename already exists
                     the number will be increased as  much as needed so
                     the file does not exist in the given location.
                     The filename can consists a path and is expanded
                     with ~ and environment variables.
    :type filename: string
    :rtype: string
    """
    location = path_expand(filename)
    n = 0
    found = True
    backup = None
    while found:
        n += 1
        backup = "{0}.bak.{1}".format(location, n)
        found = os.path.isfile(backup)
    return backup


def auto_create_version(class_name, version, filename="__init__.py"):
    """
    creates a version number in the __init__.py file.
    it can be accessed with __version__
    :param class_name:
    :param version:
    :param filename:
    :return:
    """
    version_filename = Path(
        "{classname}/{filename}".format(classname=class_name,
                                        filename=filename))
    with open(version_filename, "r") as f:
        content = f.read()

    if content != '__version__ = "{0}"'.format(version):
        banner("Updating version to {0}".format(version))
        with open(version_filename, "w") as text_file:
            text_file.write('__version__ = "{0:s}"'.format(version))


def auto_create_requirements(requirements):
    """

    creates a requirement.txt file form the requirements in the list. If the file
    exists, it get changed only if the
    requirements in the list are different from the existing file

    :param requirements: the requirements in a list
    """
    banner("Creating requirements.txt file")
    try:
        with open("requirements.txt", "r") as f:
            file_content = f.read()
    except:  # noqa: E722
        file_content = ""

    setup_requirements = '\n'.join(requirements)

    if setup_requirements != file_content:
        with open("requirements.txt", "w") as text_file:
            text_file.write(setup_requirements)


def copy_files(files_glob, source_dir, dest_dir):
    """
    copies the files to the destination

    :param files_glob: `*.yaml`
    :param source_dir: source directory
    :param dest_dir: destination directory

    """

    files = glob.iglob(os.path.join(source_dir, files_glob))
    for filename in files:
        if os.path.isfile(filename):
            shutil.copy2(filename, dest_dir)


def readfile(filename, mode='r'):
    """
    returns the content of a file
    :param filename: the filename
    :return:
    """
    if mode != 'r' and mode != 'rb':
        Console.error(f"incorrect mode : expected 'r' or 'rb' given {mode}")
    else:
        with open(path_expand(filename), mode) as f:
            content = f.read()
            f.close()
        return content


def writefile(filename, content):
    """
    writes the content into the file
    :param filename: the filename
    :param content: teh content
    :return:
    """
    with open(path_expand(filename), 'w') as outfile:
        outfile.write(content)
        outfile.truncate()


def writefd(filename, content, mode='w', flags=os.O_RDWR | os.O_CREAT, mask=0o600):
    """
    writes the content into the file and control permissions
    :param filename: the full or relative path to the filename
    :param content: the content being written
    :param mode: the write mode ('w') or write bytes mode ('wb')
    :param flags: the os flags that determine the permissions for the file
    :param mask: the mask that the permissions will be applied to
    """
    if mode != 'w' and mode != 'wb':
        Console.error(f"incorrect mode : expected 'w' or 'wb' given {mode}")

    with os.fdopen(os.open(filename, flags, mask), mode) as outfile:
        outfile.write(content)
        outfile.truncate()
        outfile.close()


def sudo_readfile(filename, split=True, trim=False):
    """
    Reads the content of the file as sudo and returns the result

    :param filename: the filename
    :type filename: str
    :param split: uf true returns a list of lines
    :type split: bool
    :param trim: trim trailing whitespace. This is useful to
                 prevent empty string entries when splitting by '\n'
    :type trim: bool
    :return: the content
    :rtype: str or list
    """
    result = subprocess.getoutput(f"sudo cat {filename}")

    if trim:
        result = result.rstrip()

    if split:
        result = result.split('\n')

    return result


# Reference: http://interactivepython.org/runestone/static/everyday/2013/01/3_password.html
def generate_password(length=8, lower=True, upper=True, number=True):
    """
    generates a simple password. We should not really use this in production.
    :param length: the length of the password
    :param lower: True of lower case characters are allowed
    :param upper: True if upper case characters are allowed
    :param number: True if numbers are allowed
    :return:
    """
    lletters = "abcdefghijklmnopqrstuvwxyz"
    uletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    # This doesn't guarantee both lower and upper cases will show up
    alphabet = lletters + uletters
    digit = "0123456789"
    mypw = ""

    def _random_character(texts):
        return texts[random.randrange(len(texts))]

    if not lower:
        alphabet = uletters
    elif not upper:
        alphabet = lletters

    for i in range(length):
        # last half length will be filled with numbers
        if number and i >= int(length / 2):
            mypw = mypw + _random_character(digit)
        else:
            mypw = mypw + _random_character(alphabet)
    return mypw


def str_bool(value):
    return str(value).lower() in ['yes', '1', 'y', 'true', 't']


def get_password(prompt):
    from cloudmesh.common.systeminfo import os_is_windows
    try:
        if os_is_windows() and is_gitbash():
            continuing = True
            while continuing:

                sys.stdout.write(prompt)
                sys.stdout.flush()
                subprocess.check_call(["stty", "-echo"])
                password = input()
                subprocess.check_call(["stty", "echo"])
                sys.stdout.write('Please retype the password:\n')
                sys.stdout.flush()
                subprocess.check_call(["stty", "-echo"])
                password2 = input()
                subprocess.check_call(["stty", "echo"])
                if password == password2:
                    continuing = False
                else:
                    Console.error('Passwords do not match\n')
            return password
        else:
            continuing = True
            while continuing:
                password = getpass(prompt)
                password2 = getpass('Please retype the password:\n')
                if password == password2:
                    continuing = False
                else:
                    Console.error('Passwords do not match\n')
            return password
    except KeyboardInterrupt:
        # Console.error('Detected Ctrl + C. Quitting...')
        if is_gitbash():
            subprocess.check_call(["stty", "echo"])
        raise ValueError('Detected Ctrl + C. Quitting...')



import pydoc

c = pydoc.locate("cloudmesh")

print(dir(c))
print(c.__package__)
print()
print(c.__spec__)

print()
print(c.__path__)
print(c.__name__)
print(c.__loader__)
print(c.__file__)
print(c.__cached__)

print()
print('\n'.join(c.__path__))


def inheritors(klass):
    """
    returns the inheritors of a class if it is loaded

    :return: a set of classes
    """
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


# d = pydoc.locate("cloudmesh.abstract.ComputeNodeABC")

# print(inheritors(d))


def get_all_subclasses(cls):
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


print(dir(c))



"""
common namespace.
"""
import pkg_resources

pkg_resources.declare_namespace(__name__)

from pprint import pprint
import os
import shutil
from cloudmesh.common.util import path_expand
from cloudmesh.common.util import yn_choice
from cloudmesh.common.util import readfile
from cloudmesh.common.util import writefile
from cloudmesh.common.util import banner
from cloudmesh.common.console import Console
from cloudmesh.common.Shell import Shell
from cloudmesh.common.debug import VERBOSE
from cloudmesh.common.dotdict import dotdict
from cloudmesh.common.FlatDict import FlatDict
from cloudmesh.common.variables import Variables

version = '4.3.149'



from hostlist import expand_hostlist
from cloudmesh.common.dotdict import dotdict


class Parameter(object):

    @staticmethod
    def parse(arguments, **kwargs):
        """
        parses arguments based on their kind and aplies Parameter.expand or
        Parameter.arguments.dict. It is defined by specifying the name in arguments,
        followed by the keyword, expand,or dict. Example

        kind={'parameter': 'expand', 'experiment': 'dict', 'COMMAND': "str'}

        arguments = parse(arguments, **kind)
        or
        arguments = parse(parameter='expand', experiment='dict', COMMAND='str')

        here the it will be applied
           arguments.parameter = Parameter.expand(arguments.paramete)
           arguments.experiment = Parameeter.arguments_to_dict(arguments.experiment)

        :param arguments: the dict or doctdict with the arguments
        :type arguments: dict or doctdict
        :param kind: dict with the names of the arguments that need to be remapped
        :type kind: dict
        :return: a doctdict with parsed and remapped arguments
        :rtype: dotdict
        """
        result = dotdict(arguments)
        if kwargs is not None:
            for k, value in kwargs.items():
                if value == 'expand':
                    result[k] = Parameter.expand(result[k])
                elif value == 'dict':
                    result[k] = Parameter.expand(result[k])
                    result[k] = Parameter.arguments_to_dict(result[k])
                elif value == 'str':
                    result[k] = ' '.join(result[k])
        return result

    @staticmethod
    def _expand(values):
        """
        given a string of the form "a,g-h,k,x-z"
        expand it tl a list with all characters between the - being expanded

        :param values: string of the form "a,g-h,k,x-z"
        :return:
        """
        if "," in values:
            found = values.split(",")
        elif "-" in values:
            _from, _to = values.split('-')
            upper = [chr(x) for x in range(65, 91)]
            lower = [chr(x) for x in range(97, 123)]
            all = upper + lower
            i_start = all.index(_from)
            i_end = all.index(_to)
            found = all[i_start:i_end + 1]  # not sure why there is +1
        else:
            found = [values]
        return found

    @classmethod
    def expand_string(cls, parameter):
        """
        can expand strings, but only allows either, or - in [] not mixed

        :param parameter: string of the form prefix[a,g-h,k,x-z]postfix
        :return:
        """
        print("O", parameter)

        if "[" not in parameter and "-" not in parameter and "," in parameter:
            return parameter.split(",")

        if "[" not in parameter and "-" not in parameter and "," not in parameter:
            return [parameter]

        elif "[" in parameter:

            prefix, found = parameter.split("[", 1)
            found, postfix = found.split("]", 1)

            found = Parameter._expand(found)

            expand = []
            for x in found:
                expand = expand + Parameter._expand(x)

            result = [f"{prefix}{x}{postfix}" for x in expand]
            return result

        return [parameter]

    @classmethod
    def expand(cls, parameter, allow_duplicates=False, sort=False, sep=":"):
        """
        Parameter.expand("a[0-1]")  -> ["a0", "a1"]
        Content sensitive : expansion
        Parameter.expand("local:a0,a1")  -> ["local:a0", "local:a1"]
        instead of
        Parameter.expand("local:[a0,a1]")  -> ["local:a0", "local:a1"]

        :param parameter:
        :param allow_duplicates:
        :param sort:
        :return:
        """
        if type(parameter) == list or parameter is None:
            return parameter

        parameters = list(expand_hostlist(parameter,
                                          allow_duplicates=False,
                                          sort=False))

        results = [t.split(sep, 1) for t in parameters]
        merge = []

        for entry in results:
            merge = merge + entry

        if len(merge) == len(parameters):
            return parameters

        elif len(merge) == len(parameters) + 1 and len(results[0]) == 2:

            prefix = results[0][0]
            _results = []
            for i in range(1, len(parameters)):
                parameters[i] = f"{prefix}:{parameters[i]}"

            return parameters

        else:

            return parameters

    @staticmethod
    def find(name, *dicts):
        """
        Finds the value for the key name in multiple dicts

        :param name: the key to find
        :param dicts: the list of dicts
        :return:
        """
        for d in dicts:
            if type(d) == str:
                return d
            elif name in d and d[name] is not None:
                return d[name]

        return None

    @staticmethod
    def find_bool(name, *dicts):
        """
        Finds the value for the key name in multiple dicts

        :param name: the key to find
        :param dicts: the list of dicts
        :return:
        """
        value = False

        for d in dicts:
            if type(d) == str:
                value = d == 'True'
            elif name in d:
                value = d[name]
            if type(value) == str:
                value = value == 'True'

            if value:
                return True

        return False

    @staticmethod
    def arguments_to_dict(arguments):
        """
        converts a string of the form "a=1,b=2" to a dict
        {"a":"1", "b":"2"}
        all values are strings

        :param arguments: the argument string
        :return: a dic of argument and values
        """
        if arguments is None or len(arguments) == 0:
            return None
        parameters = {}
        for argument in Parameter.expand(arguments):
            key, value = argument.split("=", 1)
            parameters[key] = value
        return parameters

    @staticmethod
    def separate(text, sep=":"):
        if sep in text:
            return text.split(sep, 1)
        else:
            return None, text



from pprint import pprint


class DictList(dict):
    """
    A class to convert lists of dicts to dicts.

    Example:

        data = [
        {"name": "vm1", "status": "on"},
        {"name": "vm2", "status": "on"},
        {"name": "vm3", "status": "on"},
        {"name": "vm4", "status": "on"},
        ]

        d = DictList(data)
        pprint (d)

        # {'vm1': {'name': 'vm1', 'status': 'on', 'x': 0},
        #  'vm2': {'name': 'vm2', 'status': 'on', 'x': 1},
        #  'vm3': {'name': 'vm3', 'status': 'on', 'x': 2},
        #  'vm4': {'name': 'vm4', 'status': 'on', 'x': 3}}

        print (d['vm1'])
        # {'name': 'vm1', 'status': 'on', 'x': 0}
        #
        print (d.list())
        # [{'name': 'vm1', 'status': 'on', 'x': 0},
        #  {'name': 'vm2', 'status': 'on', 'x': 1},
        #  {'name': 'vm3', 'status': 'on', 'x': 2},
        #  {'name': 'vm4', 'status': 'on', 'x': 3}]


    """

    def __init__(self, entries=None, key='name', position='x'):
        """
        initializes the DotDict List

        :param entries: a list of dict
        :param key: a key that is used as name within the dict
        :param position: the name of a key that stores the order
        """

        if type(entries) == list:
            counter = 0
            for entry in entries:
                entry[position] = counter
                self[entry[key]] = entry
                counter = counter + 1
        elif type(entries) == dict:
            # noinspection PyUnusedLocal
            self = entries
        else:
            raise ValueError("type not supported")

    def list(self):
        """
        Lists the entries

        :return:
        """
        return list(self.values())


if __name__ == "__main__":
    data = [
        {"name": "vm1", "status": "on"},
        {"name": "vm2", "status": "on"},
        {"name": "vm3", "status": "on"},
        {"name": "vm4", "status": "on"},
    ]

    d = DictList(data)
    pprint(d)

    # {'vm1': {'name': 'vm1', 'status': 'on', 'x': 0},
    #  'vm2': {'name': 'vm2', 'status': 'on', 'x': 1},
    #  'vm3': {'name': 'vm3', 'status': 'on', 'x': 2},
    #  'vm4': {'name': 'vm4', 'status': 'on', 'x': 3}}

    print(d['vm1'])
    # {'name': 'vm1', 'status': 'on', 'x': 0}
    #
    print(d.list())
    # [{'name': 'vm1', 'status': 'on', 'x': 0},
    #  {'name': 'vm2', 'status': 'on', 'x': 1},
    #  {'name': 'vm3', 'status': 'on', 'x': 2},
    #  {'name': 'vm4', 'status': 'on', 'x': 3}]
